/*
 * Standard LIBC function support for BSD sockets.
 *
 * PART 1:
 *  - stubs for standard libc functions
 *    that take a file descriptor as an argument, and/or trampolines for 
 *    such functions so that they work with sockets too (and do so
 *    without polluting the libc namespace).
 *  - interfaces with the DJGPP 'File System Extensions'.
 *    All the functions in this part are static.
 * PART 2:
 *  - Some public functions that ...
 *    a) _are_ POSIX 
 *    b) and _are_not_ BSD socket API 
 *    c) and _are_not_ in the standard dosish libcs. eg fcntl().
 *      [note: ioctl() _is_ available in the TurboC libs. Stubbed in part 1]
 * 
 * Note that although this file is fairly large, the amount of code emitted
 * is not very much since a) a great deal of it is long-winded comments like
 * this one :) and b) the bulk of it is relevant only to compilers/libs
 * other than the one you use. :)
 *
*/

/* things that need to be done before ANYTHING is included */
#if defined(__WATCOMC__)
  #undef TEST_PROG /* test must be in another file since we redef symbols */
  #define close __close  /* close() and fclose() call __close() */
  #include <io.h>
  #ifdef __cplusplus
  extern "C" {
  #endif
  extern int __IOMode(int fd);
  extern void __SetIOMode(int fd, int mode);
  #ifdef __cplusplus
  }
  #endif
#elif defined(__DJGPP__)
  #include <libc/dosio.h> /* char *__file_handle_modes */
#elif defined(__TURBOC__)
  #undef TEST_PROG /* test must be in another file since we redef symbols */
  #define read  ___cooked_read
  #define write ___cooked_write
  #define close ___cooked_close
  #include <io.h>
  #undef read
  #undef write
  #undef close
  #define read  _read
  #define write _write
  #define close _close
  /* extern int _nfile; */
  #ifdef __cplusplus
  extern "C" int _openfd[];
  #else
  extern int _openfd[];
  #endif
#endif

#include "target.h"   /* FAR */
#include "sofdmux.h"  /* __sock_fd_*(), __fsext_setup() */
#include "bsderrno.h" /* bsd error numbers */
#include "syscall.h"  /* sas_*() */

#define IS_FD_NUMBER_VALID(__fd) ((__fd & ~0xff) == 0)

/* ===================================================================== */
/*  PART 1: libc stubs and support for DJGPP's "File System extension"s  */
/* ===================================================================== */
/*
 * trampoline close(), read(), write(), dup(), dup2(), lseek()
 * [fcntl()], [ioctl()], _exit() library entry points. 
 * By proxy chsize()/ftruncate(), tell(), filelength() are covered
 * too since these all depend on lseek().
 *
 * _exit() is trampolined in order to have a consistant means of calling
 * destructors that need to be called even if the application is ending due
 * to a direct call to _exit() [fatal error, abort(), SIGSEGV, etc].
 * No more than 32 functions can be registered with the fsext_atterminate() 
 * (clever name, eh?:) function. FWIW: Watcom and MSC have a category of
 * destructors that may be called from _exit(), but DJGPP and TURBOC do not.
 *
 * No redirection is done for ..
 *  - fstat() because the stat structure's values are acceptable
 *    (same as fstat for 'nul:'), except for not having the S_ISSOCK() bit
 *    set, which we can't set because the standard headers don't #define it.
 *  - fchmod() - no DOS library appears to support this (kinda tough on DOS).
 *  - lock()/unlock()/locking() - DOS/SHARE.EXE will fail the call + EBADF.
 *  - chsize()/ftruncate(), tell(), filelength() because these all depend
 *    on lseek() which we fail with ESPIPE for sockets. (DOS itself will 
 *    not fail seeks on character devices)
 *  - umask()/setmode() - inherant support for libcs that have sensible
 *    file mode logic. (that limitation summarily excludes DJGPP ;).
 *
 * When building for DJGPP, most functions are handled by the 'File System 
 * Extension' subsystem and don't need to be trampolined. Some functions
 * DO need to be trampolined EVEN when supported by FSEXT:
 * - write() has broken O_TEXT handling when dealing with an fd that is
 *   associated with an FSEXT. See comments at __stub_write().
 * - dup() and particurly dup2() need to do pre- and post- processing for 
 *   sockets AND also need to be able to dup()/dup2() in between.
 *   (They could theoretically re-implement the guts of dup()/dup2() or
 *   alternatively clear the fsext hook, then call dup()/dup2() recursively, 
 *   then reset the hook, but since I had them working anyway, it was a lot
 *   easier to just enable them for DJGPP than rewrite).
 * Other than for write(), no other DJGPP "incongruities" are not compensated
 * for. umask() still does nothing; setmode() is still brain dead (we
 * explicitely work around it in fcntl()); read() still returns less than 
 * the requested amount for O_TEXT; stdio will still equate full buffering 
 * with line buffering; fdopen() will still stomp on the file mode used when 
 * the file was open()ed; fdopen() still doesn't fail if the read/write mode 
 * is different from what was used at open() time; O_APPEND is still thrown
 * away after the open() [seek() still only happens at open time], etc, etc.
*/

static struct __stubdata
{
  #define ISNBUF_SIZE 8
  short cs_alias; /* used by 386 protected mode */
  char isnbuf_close[ISNBUF_SIZE];    /* non-dj only */
  char isnbuf_read[ISNBUF_SIZE];     /* non-dj only */
  char isnbuf_write[ISNBUF_SIZE];    /* non-dj only */
  char isnbuf_lseek[ISNBUF_SIZE];    /* non-dj only */
  char isnbuf_dup[ISNBUF_SIZE];      /* all */
  char isnbuf_dup2[ISNBUF_SIZE];     /* all */
  char isnbuf_ioctl[ISNBUF_SIZE];    /* turboc/bcc only */
  #if defined(__WATCOMC__)           /* all watcom stdio sits on these two */
  char isnbuf___qread[ISNBUF_SIZE];  /* they are "light" (q="q"uick maybe?) */
  char isnbuf___qwrite[ISNBUF_SIZE]; /* and small. no arg/env validation etc */
  #endif
} stubdata;

/* --------------------------------------------------------------- */

static void __stubdata_init_if_needed(void)
{
  static int my_bss_count = -1;
  if (__bss_count != my_bss_count)
  {
    memset( &stubdata, 0, sizeof(stubdata) );
    my_bss_count = __bss_count;
  }
  return;
}

/* --------------------------------------------------------------- */

#if defined(__MEDIUM__) || defined(M_I86MM)
#define CODETYPE _far
#elif defined(__COMPACT__) || defined(M_I86CM)
#define CODETYPE _near
#elif defined(__HUGE__) || defined(M_I86HM)
#define CODETYPE _far
#else
#define CODETYPE /* whatever */
#endif

static void __read_code(char *isnbuf, const void CODETYPE *libcfunc, size_t len)
{
#if defined(__DPMI32__)
  movedata ( __my_dpmi_get_cs(), (unsigned)libcfunc, 
             __my_dpmi_get_ds(), (unsigned)isnbuf, len );
#else  /* real mode targets */
  _fmemcpy( isnbuf, libcfunc, len );
#endif
}

/* --------------------------------------------------------------- */
  
static void __write_code(void CODETYPE *libcfunc, const char *isnbuf, size_t len)
{
#if defined(__DPMI32__)
  _disable();
  movedata ( __my_dpmi_get_ds(), (unsigned)isnbuf, 
             stubdata.cs_alias, (unsigned)libcfunc, len );
  _enable();
#else
  _disable();
  _fmemcpy( libcfunc, isnbuf, len );
  _enable();
#endif
}

/* --------------------------------------------------------------- */

static int restore_libcfunc(void CODETYPE *libcfunc, const char *isnbuf, int lock_state)
{
  __write_code(libcfunc, isnbuf, ISNBUF_SIZE);
  if (lock_state < 0)
  {
    #if defined(__DPMI32__)
    __my_dpmi_unlock_code (libcfunc, ISNBUF_SIZE);
    #endif
  }
  return 0;
}

/* --------------------------------------------------------------- */

static int redirect_libcfunc(char *sav_isnbuf, void CODETYPE *libcfunc, 
                             const void CODETYPE *ourfunc, int lock_state )
{
  unsigned int szi = sizeof(int), szp = sizeof(void CODETYPE *);
  char new_isnbuf[ISNBUF_SIZE];
  unsigned long offset;

  __read_code( sav_isnbuf, libcfunc, ISNBUF_SIZE );
  if (lock_state > 0) /* note: read data first to ensure that */
  {                   /* it is in physical memory */
    /* must lock to prevent page-in from the .exe image */
    #if defined(__DPMI32__)
    __my_dpmi_lock_code (libcfunc, ISNBUF_SIZE);  
    #endif
  }

  memcpy( new_isnbuf, sav_isnbuf, ISNBUF_SIZE );
  if (szi == 4) /* 386 instruction set */
  {
    offset = ((const char *)ourfunc) - ((const char *)libcfunc);
    new_isnbuf[0] = 0xe9;    /* "jmp" (relative) */
    *((unsigned long *)(&new_isnbuf[1])) = offset - 5; /* 5==sizeof('jmp xyz') */
  }
  else if (szp == 2) /* small code */
  {
    offset = ((const char *)ourfunc) - ((const char *)libcfunc);
    new_isnbuf[0] = 0xe9;    /* "jmp" (intra segment) */
    *((unsigned short *)(&new_isnbuf[1])) = ((short)(offset - 3));
  }
  else /* large code */
  {
    offset = ((unsigned long)ourfunc);
    new_isnbuf[0] = 0xEA;    /* jmp absolute */
    *((unsigned long *)(&new_isnbuf[1])) = offset;
  }
  __write_code( libcfunc, new_isnbuf, ISNBUF_SIZE );

  return 0;
}

#undef CODETYPE

/* --------------------------------------------------------------- */

#define REDIRECT_LIBCFUNC(__name) \
        redirect_libcfunc( stubdata.isnbuf_##__name , \
                           __name, \
                           __stub_##__name , 0 )
#define RESTORE_LIBCFUNC(__name) \
        restore_libcfunc(__name, stubdata.isnbuf_##__name , 0 )

#if !defined(DEBUG_FSEXT)
#define FSEXT_DPRINTF(x) /* nothing */
#define FSEXT_DDUMP(__ctx, __pkt, __len ) /* nothing */
#else
#include "printk.h"
#define FSEXT_DPRINTF(x) _printk x
static void FSEXT_DDUMP( const char *label, const char *apacket,
                                            unsigned int alen )
{
  unsigned int i;
  for (i = 0; i < alen; i += 16)
  {
    char buffer[128];
    char *p, *q; unsigned int n;
    _sprintk(buffer,"\r%s %04x: ", label, i );
    q = 48 + (p = &buffer[strlen(buffer)]);
    for (n = 0; n < 16; n++)
    {
      unsigned int c = ' ';
      p[0] = p[1] = ' ';
      if (( n + i ) < alen )
      {
        static const char *tox="0123456789abcdef";
        c = (((unsigned int)apacket[n+i]) & 0xff);
        p[0] = (char)tox[c>>4];
        p[1] = (char)tox[c&0x0f];
        if (!isprint(c) || /*isctrl(c)*/ c=='\r' || c=='\n' || c=='\t')
          c = '.';
      }
      p+=2;
      *p++ = ' ';
      *q++ = (char)c;
    }
    *q++ = '\n';
    *q = '\0';
    _printk("%s",buffer);
  }
  return;
}
#endif

/* --------------------------------------------------------------- */

#if !defined(__DJGPP__) /* DJGPP has fs extensions */
static int __stub_close(int fd)
{
  int rc = -1;
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  FSEXT_DPRINTF(("__stub_close(%d) socket?=>%p\n",fd,socket));
  if (RESTORE_LIBCFUNC(close)!=0)
    errno = EFAULT;
  else
  {
    if (!socket)
      rc = close(fd);
    else if ((rc = bsdsys_close(socket))==0) 
      __sock_fd_free(fd);
    REDIRECT_LIBCFUNC(close);
  }
  FSEXT_DPRINTF(("__stub_close(%d)=>%d [err='%s']\n",fd,rc,
                              ((rc!=-1)?("none"):(strerror(errno))) ));
  return rc;
}
#endif

/* --------------------------------------------------------------- */

#if !defined(__DJGPP__) /* DJGPP has fs extensions */
static long __stub_lseek(int fd, long offset, int whence)
{
  long rc = -1L;
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  FSEXT_DPRINTF(("__stub_lseek(%d,%ld,%d) socket?=>%p\n",
                                          fd,offset,whence,socket));
  if (socket)
  {
    #if defined(ESPIPE)
    errno = ESPIPE;
    #else
    errno = EFAULT;
    #endif
  }    
  else if (RESTORE_LIBCFUNC(lseek)!=0)
    errno = EFAULT;
  else
  {
    rc = lseek(fd,offset,whence);
    REDIRECT_LIBCFUNC(lseek);
  }
  FSEXT_DPRINTF(("__stub_lseek(%d)=>%d [err='%s']\n",fd,rc,
                              ((rc!=-1L)?("none"):(strerror(errno))) ));
  return rc;
}
#endif

/* --------------------------------------------------------------- */

/*
** Following the FSExt paradigm, what we want to trampoline for read/write is
** the libc function that does the raw DOS Int21h I/O (ie, no translation)
** and which is called by higher level functions that do translation.
**
** - TurboC:  write()/fwrite() both call _write() which only does raw writes.
**            read()/fread() both call _read() which only does raw reads.
**            This is exactly what we want, so no special handling needed.
**            (_openfd[fd] might need the O_DEVICE flag cleared. Don't know)
**
** - WatcomC: write() and read() do cooking, and do all DOS I/O themselves, 
**            so we have to emulate write's/read's O_TEXT handling ourselves.
**            fwrite()/fread() however call __qwrite()/__qread() for raw I/O.
**            So, we need to stub those too.
**
** - DJGPP:   In theory, DJGPP is just like TurboC, with _write()/_read()
**            calling the FSExtension hook. However, current DJGPP (2.03) 
**            has a bug in write() - for O_TEXT it only cooks the buffer when 
**            writing to a dos file (yes, write() does int21h I/O directly
**            for O_TEXT mode). FS Extensions get called with unprocessed 
**            data, even if the file is O_TEXT.
**            We trampoline write() instead of doing the xlation in the fsext 
**            handler because a) DJGPP might eventually fix this bug, and we 
**            don't want to be cooking the data twice, and b) stdio uses 
**            _write() and the data will in that case be pre-cooked.
*/
#if defined(__WATCOMC__)
  #define O_TEXT_NEEDS_SEND_COOKER   /* do \n->\r\n before send()ing */
  #define O_TEXT_NEEDS_RECV_UNCOOKER /* do \r\n->\n after recv()ing */
#elif defined(__DJGPP__)
  #define O_TEXT_NEEDS_SEND_COOKER   /* do \n->\r\n before send()ing */
#endif

#if defined(O_TEXT_NEEDS_SEND_COOKER) || defined(O_TEXT_NEEDS_RECV_UNCOOKER)
static int __is_text_mode(int fd) 
{
  #if defined(__DJGPP__)         /* *DONT* use setmode()! */
  return ((__file_handle_modes[fd] & (O_BINARY|O_TEXT)) == O_TEXT);
  #elif defined(__WATCOMC__)
  return ((__IOMode(fd) & _BINARY) == 0);
  #else
  #error how did you get here?
  #endif
}  
#endif

/* --------------------------------------------------------------- */

#if defined(O_TEXT_NEEDS_SEND_COOKER)
static int __send_text(struct fdmux_object FAR *socket, 
                       const char *buf, unsigned int buflen)
{
  int rc = 0;
  struct msghdr msg;
  struct iovec iov;
  memset( &msg, 0, sizeof(msg));
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;

  while (buflen)
  {
    char tmpbuf[512]; /* should prolly use mtu * x here */
    const char *obuf = (const char *)memchr(buf, '\n', buflen);
    unsigned int used;
    
    if (!obuf || (obuf - buf) > sizeof(tmpbuf))
    {
      used = (obuf ? (obuf - buf) : buflen);
     
      iov.iov_len = used;
      iov.iov_base = (char *)(*((void **)&buf));

      FSEXT_DDUMP("sendmsg A",iov.iov_base,iov.iov_len);
      used = (unsigned)bsdsys_sendmsg( socket, &msg, 0 );
      if (((int)used) == -1 || (used == 0 && rc == 0))
      {
        if (rc == 0)
          rc = -1;
        break; 
      }
  
      rc += used;
      buflen -= used;
      buf += used;

      if (used < (unsigned)iov.iov_len)
        break;
    }
    else
    {
      obuf = buf;
      used = 0;
       
      while (used < (sizeof(tmpbuf)-2) && buflen)
      {
        if (*buf == '\n')
          tmpbuf[used++] = '\r';
        tmpbuf[used++] = *buf++;
        buflen--;
      }

      iov.iov_len = used;
      iov.iov_base = &tmpbuf[0];

      FSEXT_DDUMP("sendmsg B",iov.iov_base,iov.iov_len);
      used = (unsigned)bsdsys_sendmsg( socket, &msg, 0 );
      if (((int)used) == -1 || (used == 0 && rc == 0))
      {
        if (rc == 0)
          rc = -1;
        break; 
      }

      if (used < (unsigned int)iov.iov_len)  
      { 
        unsigned int done = 0; 
        buf = obuf; 
        while (done < used)
        {
          if (*buf++ == '\n')
            done++;
          done++;
        }
        rc += buf - obuf;
        break;
      }

      rc += buf - obuf;

    } /* if needed translation */
  } /* while (buflen) */
  return rc;
}
#endif /* O_TEXT_NEEDS_SEND_COOKER */

/* --------------------------------------------------------------- */

static int __stub_write(int fd, const char *buf, unsigned int buflen)
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  int rc = -1;

  FSEXT_DPRINTF(("__stub_write(%d,%p,%u) socket?=>%p\n",fd,buf,buflen,socket));

  if (buflen == 0) /* so says posix */
    rc = 0;
  else if ((buflen + 1) < 2)
    errno = EINVAL;
#if defined(O_TEXT_NEEDS_SEND_COOKER)
  else if (socket && __is_text_mode(fd))
    rc = __send_text(socket, buf, buflen);
#endif
  else if (socket)
  {
    FSEXT_DDUMP("send()",buf,buflen);
    rc = send( fd, buf, buflen, 0 );
  }
  else if (RESTORE_LIBCFUNC(write)!=0)
    errno = EFAULT;
  else
  {
    rc = write( fd, buf, buflen);
    REDIRECT_LIBCFUNC(write);
  }

  FSEXT_DPRINTF(("__stub_write(%d)=>%d [err='%s']\n",fd,rc,
                              ((rc!=-1)?("none"):(strerror(errno))) ));
  return rc;
}

/* --------------------------------------------------------------- */

#if defined(O_TEXT_NEEDS_RECV_UNCOOKER)
static int __recv_text(int fd, char *buf, unsigned int buflen)
{
  int rc = -1; 
  while (buflen)
  {
    char *in, *start;
    unsigned int used = recv( fd, buf, buflen, 0 );
    if (((int)used) == -1) 
      break;
    start = in = buf;
    while (used)
    {
      if (*in == 0x1a) /* EOF */
        break;
      if (*in != '\r')
        *buf++ = *in;
      in++;
      used--;   
    }
    if (rc == -1)
      rc = 0;
    rc += (buf - start);
    if ((in - start) < buflen) /* read less than wanted */
      break;                   /* or used > 0 (EOF) */
    buflen -= (buf - start);  
  }
  return rc;
}
#endif /* O_TEXT_NEEDS_RECV_UNCOOKER */

/* --------------------------------------------------------------- */

#if !defined(__DJGPP__) /* DJGPP has fs extensions */
static int __stub_read(int fd, char *buf, unsigned int buflen)
{
  int rc = -1;
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  FSEXT_DPRINTF(("__stub_read(%d,%p,%u) socket?=>%p\n",fd,buf,buflen,socket));

  if (buflen == 0) /*in case you've ever wondered how to do an EBADF test :)*/
  {
    if (socket || !__sock_fd_isEBADF(fd))
      rc = 0;
    else
      errno = EBADF;
  }
#if defined(O_TEXT_NEEDS_RECV_UNCOOKER)
  else if (socket && __is_text_mode(fd))
    rc = __recv_text(fd, buf, buflen);
#endif
  else if (socket)
    rc = recv( fd, buf, buflen, 0 );
  else if (RESTORE_LIBCFUNC(read)!=0)
    errno = EFAULT;
  else
  {
    rc = read( fd, buf, buflen);
    REDIRECT_LIBCFUNC(read);
  }

  if (rc >= 0 && socket) {
    FSEXT_DDUMP("__stub_read",buf,buflen);
  }
  FSEXT_DPRINTF(("__stub_read(%d)=>%d [err='%s']\n",fd,rc,
                              ((rc!=-1)?("none"):(strerror(errno))) ));
  return rc;
}
#endif

/* --------------------------------------------------------------- */

#if defined(__WATCOMC__)
#ifdef __cplusplus
extern "C" {
#endif
extern int __qwrite( int handle, const void *buffer, unsigned len );
extern int __qread( int handle, char *buffer, unsigned len );
#ifdef __cplusplus
}
#endif
#endif

#if defined(__WATCOMC__)
static int __stub___qwrite(int fd, const char *buf, unsigned int buflen)
{
  int rc = -1;
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);

  FSEXT_DPRINTF(("__stub___qwrite(%d,'%p',%u) socket?=>%p\n",fd,buf,buflen,socket));
  FSEXT_DDUMP("__stub__qwrite",buf,buflen);

  if (socket)
    rc = send( fd, buf, buflen, 0 );
  else if (RESTORE_LIBCFUNC(__qwrite)!=0)
    errno = EFAULT;
  else
  {
    rc = __qwrite(fd,buf,buflen);
    REDIRECT_LIBCFUNC(__qwrite);
  }
  FSEXT_DPRINTF(("__stub__qwrite(%d)=>%d [err='%s']\n",fd,rc,
                              ((rc!=-1)?("none"):(strerror(errno))) ));
  return rc;
}
#endif /* ___WATCOMC__ */

/* --------------------------------------------------------------- */

#if defined(__WATCOMC__)
static int __stub___qread(int fd, char *buf, unsigned int buflen)
{
  int rc = -1;
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  FSEXT_DPRINTF(("__stub___qread(%d,%p,%u) socket?=>%p\n",fd,buf,buflen,socket));
  if (socket)
    rc = recv( fd, buf, buflen, 0 );
  else if (RESTORE_LIBCFUNC(__qread)!=0)
    errno = EFAULT;
  else
  {
    rc = __qread( fd, buf, buflen);
    REDIRECT_LIBCFUNC(__qread);
  }
  FSEXT_DPRINTF(("__stub___qread(%d)=>%d [err='%s']\n",fd,rc,
                              ((rc!=-1)?("none"):(strerror(errno))) ));
  return rc;
}
#endif /* ___WATCOMC__ */

/* --------------------------------------------------------------- */

#if defined(__TURBOC__) /* for both borland and tc */
/* turboc's ioctl() function is a frontend to DOS' Int21h/ah=0x44 call.
** 'func' is the AL value for the call (DOS supports 0x00...0x10)
** optional 3rd and 4th args are: void _FAR * argdx, int argcx
**
** NOTE that 'func' is an 'int', not a long. Consequently, there has
** to be some hackery done in sys/socket.h in order to have an ioctl()
** function that takes a long.
*/ 
static int __stub_ioctl (int handle, int func, ...)
{ 
  int rc = -1;        
  va_list va;
  if ((func & 0xff00) == 0) /* its a libc [DOS] ioctl() */
  {                         /* (actually only 0x00...0x10) */  
    FSEXT_DPRINTF(("[LIBC call] __stub_ioctl(%d,0x%04x)\n",fd,func));
    if (RESTORE_LIBCFUNC(ioctl)!=0)
      errno = EFAULT;
    else
    {
      void _FAR *argdx;
      int argcx;
      va_start (va, func);
      argdx = va_arg (va, void _FAR *);
      argcx = va_arg (va, int);
      va_end (va);
      rc = ioctl( handle, func, argdx, argcx );
      REDIRECT_LIBCFUNC(ioctl);
    }
    FSEXT_DPRINTF(("[LIBC call] __stub_ioctl(%d)=>%d (err=%s)\n",fd,rc, 
                                ((rc!=-1)?("none"):(strerror(errno))) ));
  }
  else
  {
    struct fdmux_object FAR *socket = __sock_fd_lookup(handle);
    FSEXT_DPRINTF(("__stub_ioctl(%d) socket?=>%p\n",fd,socket));
    if (!socket)
      errno = ENOTSOCK;
    else
    {
      long cmd;
      char *argp;
      va_start (va, handle);
      cmd = va_arg (va, long);
      argp = va_arg (va, char *);
      va_end (va);
      rc = bsdsys_ioctl(socket, cmd, argp);
    }      
    FSEXT_DPRINTF(("__stub_ioctl(%d)=>%d (err=%s)\n",fd,rc, 
                                ((rc!=-1)?("none"):(strerror(errno))) ));
  }
  return rc;
}
#endif /* TURBOC */

/* --------------------------------------------------------------- */

/* caution: __stup_dup() and __stub_dup2() is hairy code. 
** Do not modify unless you understand what is being called from
** functions called from here.
*/ 
static int __stub_dup(int fd) /* for all libraries */ 
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  int newfd = -1;

  FSEXT_DPRINTF(("__stub_dup called for %s\n",(socket?"socket":"non-socket")));

  if (RESTORE_LIBCFUNC(dup)!=0)
    errno = EFAULT;
  else
  {
    newfd = dup(fd);
    REDIRECT_LIBCFUNC(dup);
  }
  /* caution: don't try to "optimize" this - 
  ** __sock_fd_alloc() might call dup() 
  */
  if (newfd != -1 && socket)
  {
    socket = bsdsys_dup(socket);
    if (!socket)
    {
      close(newfd);
      newfd = -1;
    }
    else
    {
      __sock_fd_assoc(newfd, socket);
      #if defined(__DJGPP__)
      __FSEXT_set_function(newfd, __djgpp_fsext_demux);
      #endif
    }
  }
  FSEXT_DPRINTF(("__stub_dup result=> %d\n",newfd));
  return newfd;
}

/* ------------------------------------------------------------------ */

/* caution: __stup_dup() and __stub_dup2() is hairy code. 
** Do not modify unless you understand what is being called from
** functions called from here.
*/ 
static int __stub_dup2(int fd, int fd2) /* for all libraries */
{
  int rc = -1;
  FSEXT_DPRINTF(("__stub_dup2(%d,%d)\n",fd,fd2));
  if (__sock_fd_isEBADF(fd)) /* applies to both libc/dos and skts */
    errno = EBADF; /* handle1 is out of range or not open */
  else if (fd == fd2)
    rc = 0;  
  else if (!IS_FD_NUMBER_VALID(fd2)) /* applies to both libc/dos and skts */
    errno = EBADF; /* handle2 is out of range */
  else
  {    
    struct fdmux_object FAR *newsock = __sock_fd_lookup(fd);
    rc = 0; /* assume fd isn't a socket */
    if (newsock)
    {
      newsock = bsdsys_dup(newsock);
      if (!newsock)
        rc = -1; /* errno is already set */    
    }
    if (rc == 0) /* 'fd' either isn't a socket, or bsdsys_dup() was ok */
    {
      if (RESTORE_LIBCFUNC(dup2)!=0)
      {
        errno = EFAULT;
        rc = -1;
      }  
      else
      {
        /* get the socket associated with 'fd2' - If dup2 calls close()
        ** (which we are trapping), then the association will be unlinked
        ** automatically, otherwise we need to unlink the association manually.
        */
        struct fdmux_object FAR *oldsock = __sock_fd_lookup(fd2);

        rc = dup2(fd, fd2); /* might call close() - we check that below */
        REDIRECT_LIBCFUNC(dup2);

        /* was a socket && (dup2 succeeded || pre-dup close) */
        if (oldsock && (rc == 0 || __sock_fd_isEBADF(fd2)))
        {
          /* if 'fd2' is still in the fdlookup table, then fd2 wasn't closed
          ** using close(), but via some other libc internal mechanism or 
          ** left to DOS to close.
          */ 
          if (oldsock == __sock_fd_lookup(fd2)) /* its still there? */
          {                                     /* libc didn't use close() */
            /* we can't call close() here because fd2==fd already */
            bsdsys_close(oldsock); /* decrements fd_duped if necessary */
            __sock_fd_assoc( fd2, (struct fdmux_object FAR *)0 );
            #if defined(__DJGPP__)
            __FSEXT_set_function(fd2, NULL);
            #endif
          }
        } /* was a socket && (dup2 succeeded || pre-dup close) */
      } /* RESTORE_LIBCFUNC() suceeded */

      if (newsock) /* fd is socket, and bsdsys_dup() had succeeded */
      {
        if (rc != 0) /* the dup2 failed */
          bsdsys_close(newsock); /* decrement the dup count */
        else         /* the dup2 suceeded */
        {
          __sock_fd_assoc( fd2, newsock );
          #if defined(__DJGPP__)
          __FSEXT_set_function(fd2, __djgpp_fsext_demux);
          #endif
        }
      }
    }
  }
  FSEXT_DPRINTF(("__stub_dup2 result=>%d\n",rc));
  return rc;
}

/* --------------------------------------------------------------- */

#if !defined(__DJGPP__) /* DJGPP fcntl.h already has these */
#define F_DUPFD  1   /* duplicate file descriptor */
#define F_GETFD  2   /* get file descriptor flags (FD_CLOEXEC==O_NOINHERIT) */
#define F_SETFD  3   /* set file descriptor flags (FD_CLOEXEC==O_NOINHERIT) */
#define F_GETFL  4   /* get file status flags */
#define F_SETFL  5   /* set file status flags */
#define F_GETLK  6   /* get record locking info - unsupported for DOS */
#define F_SETLK  7   /* set record locking info - unsupported for DOS */
#define F_SETLKW 8   /* F_SETLK+wait if locked */

#if defined(__WATCOMC__)
   #define _O_NONBLOCK 0x4000
   #define _O_ASYNC    0x8000
#elif defined(__TURBOC__)
   #define _O_NONBLOCK 0x0008  /* undefined - but its open() time use only */
   #define _O_ASYNC    0x0400  /* O_EXCL - free after open() */
#elif defined(_MSC_VER) && defined(_O_TEMPORARY)
   #define _O_NONBLOCK _O_SHORT_LIVED  /* reuse the no-flush flag */
   #define _O_ASYNC    _O_SEQUENTIAL   /* this fits nicely too */
#else
   #error I don't know anything about this library.
#endif

#define FD_CLOEXEC  O_NOINHERIT

#define O_NONBLOCK  _O_NONBLOCK 
#define O_ASYNC     _O_ASYNC    
#define FNDELAY     O_NONBLOCK   /* old BSD compatibility */
#define FASYNC      O_ASYNC      /* old BSD compatibility */

#define F_RDLOCK  1  /* shared or read lock */
#define F_UNLOCK  2  /* unlock */
#define F_WRLOCK  3  /* exclusive or write lock */

#if defined(__BORLANDC__)
typedef unsigned long off_t;
#endif

struct flock {
         off_t l_start;  /* starting offset relative to l_whence */
         off_t l_len;    /* length - len = 0 means until end of file */
         int   l_pid;    /* lock owner (s/int/pid_t/ if available) */
         short l_type;   /* F_RDLOCK/F_WRLOCK/F_UNLOCK */
         short l_whence; /* type of l_start */
};
#endif /* !DJGPP */

/* --------------------------------------------------------------- */

/* fcntl() for both sockets and normal libc/dos descriptors */
/* __do_fcntl() is shared by djgpp fs extensions and non-djgpp libcs */
static int __do_fcntl(int fd, int cmd, int arg)
{
  #if defined(__WATCOMC__)
  #endif
  arg = arg; /* possibly unused */
  cmd = cmd; /* possibly unused */

  if (__sock_fd_isEBADF(fd))
  {
    errno = EBADF;
    return -1;
  }

#if defined(F_DUPFD)
  if (cmd == F_DUPFD)
  {
    if (arg < 0)
      arg = 0;
    while (IS_FD_NUMBER_VALID(arg))
    {
      if (__sock_fd_isEBADF(arg)) /* fd is free */
        return dup2(fd, arg);
      arg++;
    }
    errno = EMFILE;
    return -1;
  }
#endif

#if defined(F_GETFL)
  if (cmd == F_GETFL)
  {
    int rc = 0;
    struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
    #if defined(__TURBOC__)
    {
      rc |= ((O_BINARY|O_TEXT|O_APPEND) & _openfd[fd]);
    }
    #elif defined(__WATCOMC__)
    {
      int ec = __IOMode(fd);
      if (ec & _APPEND )
        rc |= O_APPEND;  
      if (ec & _BINARY)
        rc |= O_BINARY;
      else
        rc |= O_TEXT;
    }
    #elif defined(__DJGPP__)
    {
      extern char *__file_handle_modes;
      rc = __file_handle_modes[fd];
      rc &= (O_BINARY|O_TEXT);
    }
    #endif
    if (socket)
    {
      #if defined(O_NONBLOCK)
      if ((socket->so_state & SS_NBIO)!=0)
        rc |= O_NONBLOCK;
      #endif
      #if defined(O_ASYNC)
      if ((socket->so_state & SS_ASYNC)!=0)
        rc |= O_ASYNC;
      #endif
    }
    return rc;
  }
#endif

#if defined(F_SETFL)
  if (cmd == F_SETFL)
  {
    int rc = 0;
    struct fdmux_object FAR *socket = __sock_fd_lookup(fd);

    /* make a list of O_ flags that may be set */ 
    rc = O_BINARY|O_TEXT|O_APPEND;
    if (socket)
    {
      #ifdef O_NONBLOCK
      rc |= O_NONBLOCK;
      #endif
      #ifdef O_ASYNC
      rc |= O_ASYNC;
      #endif
    }
    
    if ((arg & ~rc) != 0 || /* reject an arg with any invalid bits set */
     (arg & (O_TEXT|O_BINARY)) == (O_TEXT|O_BINARY) ) /* not both together */
    {
      errno = EINVAL;
      return -1;
    }

    /* Normally, if a bit is not present in 'arg' it is cleared 
    ** on the fd, but since O_BINARY|O_TEXT bits are non-posix, 
    ** it is possible that neither are in 'arg' and we set/clear
    ** only if one or the other is there.
    */ 

    rc = 0;
    #if defined(__TURBOC__)
    {
      rc |= ((O_BINARY|O_TEXT|O_APPEND) & _openfd[fd]);
      if (arg & O_APPEND) /* set only */
        _openfd[fd] |= O_APPEND;
      if ((arg & (O_BINARY|O_TEXT))!=0)
      {
        /* Q: for sockets, do we need to do _openfd[fd] &= ~O_DEVICE
        ** in order for O_TEXT mode to work correctly? 
        */
        setmode( fd, (arg & (O_BINARY|O_TEXT)));
      }
    }
    #elif defined(__WATCOMC__)
    {
      int ec = __IOMode( fd );
      if (ec & _APPEND)
        rc |= O_APPEND;
      if (ec & _BINARY)
        rc |= O_BINARY;
      else
        rc |= O_TEXT;
      if ((arg & O_APPEND)!=0) /* set only */
        ec |= _APPEND;
      if ((arg & (O_BINARY|O_TEXT))!=0)
      {
        ec &= ~_BINARY;
        if ((arg & O_BINARY)!=0)
          ec |= _BINARY;
        setmode( fd, (arg & (O_BINARY|O_TEXT)));
      }
      __SetIOMode(fd, ec );
    }
    #elif defined(__DJGPP__)
    {
      rc |= (__file_handle_modes[fd] & (O_TEXT|O_BINARY));
      if ((arg & (O_BINARY|O_TEXT))!=0)
      {
        /* don't use setmode() for DJGPP! */ 
        __file_handle_modes[fd] &= ~(O_TEXT|O_BINARY);        
        __file_handle_modes[fd] |= (arg & (O_TEXT|O_BINARY));
      }
    }
    #endif
    if (socket)
    {
      #if defined(O_NONBLOCK)
      if ((socket->so_state & SS_NBIO)!=0)
        rc |= O_NONBLOCK;
      if ((arg & O_NONBLOCK)!=0)
        socket->so_state |= SS_NBIO;
      else
      socket->so_state &= ~SS_NBIO;
      #endif
      #if defined(O_ASYNC)
      if ((socket->so_state & SS_ASYNC)!=0)
        rc |= O_ASYNC;
      if ((arg & O_ASYNC)!=0)
        socket->so_state |= SS_ASYNC;
      else
        socket->so_state &= ~SS_ASYNC;
      #endif
    }
    return rc;
  } /* if (cmd == F_SETFL) */
#endif
#if defined(F_GETOWN)
  if (cmd == F_GETOWN)
    return 0; /* owner is root */
#endif
#if defined(F_SETOWN)
  if (cmd == F_SETOWN)
    return 0; /* previous owner :) */
#endif
#if defined(F_GETFD) 
  if (cmd == F_GETFD)
  {
    if (__sock_fd_lookup(fd))
      return 1; /* no inherit flag is set */
    return 0; /* don't really know actually */
  } 
#endif
#if defined(F_SETFD) 
  if (cmd == F_SETFD)
  {
    if (__sock_fd_lookup(fd))
      return 1; /* previous state of no-inherit flag */
    return 0; /* don't really know actually */
  }
#endif

  /* F_GETLK, F_SETLK, F_SETLKW */
  errno = EOPNOTSUPP;
  return -1;
}

/* =============================================================== */
/*                                                                 */
/* =============================================================== */

#if defined(__DJGPP__)
static int __djgpp_fsext_demux(__FSEXT_Fnumber func, int *rv, va_list _args)
{
  int    fd   = va_arg (_args, int);
  int    cmd  = va_arg (_args, int);
  int    arg  = va_arg (_args, int);
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);

  if (socket)
  {
    switch (func)
    {
      case __FSEXT_read: /* called from _read() and fread() */
      {
        *rv = recv(fd, (char*)cmd, arg, 0);
        return 1; /* handled */
      }
      case __FSEXT_write: /* called from _write(), write() and fwrite() */
      {
        *rv = send(fd, (char*)cmd, arg, 0);
        return 1;
      }
      case __FSEXT_close:
      {
        *rv = bsdsys_close(socket);
        if (*rv == 0)         /* __sock_fd_free calls close() too - but by */
          __sock_fd_free(fd); /* then the fsext hook will have been removed */
        return 1; /* handled */
      }
      case __FSEXT_ioctl:   /* ioctl (fd,cmd,...) */
      {
        *rv = bsdsys_ioctl(socket, cmd, (char*)arg);
        return 1;
      }
      case __FSEXT_fcntl:   /* 'fcntl (fd,cmd)' or 'fcntl (fd,cmd,arg)' */
      {
        *rv = __do_fcntl(fd, cmd, arg); /* static function */
        return 1;
      }
#if 0 /* unused - __sock_fd_select() is a complete replacement for select() */
      case __FSEXT_ready:
      {
        struct  timeval timeout;
        fd_set  fd_read, fd_write, fd_err;
        FD_ZERO (&fd_read);
        FD_ZERO (&fd_write);
        FD_ZERO (&fd_err);
        FD_SET (fd, &fd_read);
        FD_SET (fd, &fd_write);
        FD_SET (fd, &fd_err);
        timeout.tv_sec  = 0;
        timeout.tv_usec = 0L;
        *rv = sock_fd_select(fd+1, &fd_read, &fd_write, &fd_err, &timeout);
        if (*rv > 0)
        {
          *rv = 0;
          if (FD_ISSET(fd,&fd_read))  *rv |= __FSEXT_ready_read;
          if (FD_ISSET(fd,&fd_write)) *rv |= __FSEXT_ready_write;
          if (FD_ISSET(fd,&fd_err))   *rv |= __FSEXT_ready_error;
        }
        return 1;
      }     
#endif
      case __FSEXT_lseek:
         errno = ESPIPE;
         *rv = -1;
         return 1;
      case 12: /* __FSEXT_dup: djgpp >= 2.02 */
         break; /* never called - dup() needs special handling and is stubbed */
      case 13: /* __FSEXT_dup2: djgpp >= 2.02 */
         break; /* never called - dup2() needs special handling and is stubbed */
      case 14: /*__FSEXT_fstat: djgpp >= 2.02*/
      { 
        struct stat *statblk = (struct stat *)cmd;
        memset( statblk, 0, sizeof(*statblk));
        statblk->st_dev = -1;
        statblk->st_ino = (long)socket;
        statblk->st_mode = 0600;
        #if defined(S_IFSOCK)
        statblk->st_mode |= S_IFSOCK;  
        #elif defined(S_IFCHR)
        statblk->st_mode |= S_IFCHR;  
        #endif
        *rv = 0; 
        return 1;
      }
      default:
      {
        break;
      }
    } /* switch */
  }
  return 0; /* unhandled extension */
}
#endif /* defined(__DJGPP__) */

/* ========================================================================= */
/* PART 2 - POSIX functions that are not network-specific                    */
/* ========================================================================= */

#if !defined(__DJGPP__)
int fcntl(int fd, int cmd, ...)  /* all non-DJGPP libcs have no fcntl */
{
  int arg; va_list va;
  va_start (va, cmd);
  arg = va_arg (va, int);
  return __do_fcntl(fd, cmd, arg);
}
#endif

/* --------------------------------------------------------------- */

#if !defined(__DJGPP__) && !defined(__TURBOC__) 
/* DJGPP does this through fsext. TURBOC/BORLANDC through __stub_ioctl()
** WARNING: if you change sizeof(cmd) to int, don't forget to modify 
** the TURBOC code in section 2 of fsext.c
*/
int ioctl (int fd, long cmd, ...)
{
  int rc = -1;
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (!socket)
    errno = ENOTSOCK;
  else
  {
    va_list va;
    char *argp;
    va_start (va, cmd);
    argp = va_arg (va, char *);
    va_end (va);
    rc = bsdsys_ioctl(socket, cmd, argp);
  }
  return rc;
}
#endif

/* --------------------------------------------------------------- */

static int __iov_einval_check(const struct iovec *iov, unsigned int iovlen)
{
  int i, xferred;
  if (!iov || iovlen <= 0 || iovlen > 16)
  {
    errno = EINVAL;
    return -1;
  }
  xferred = 0;
  for (i = 0; i < iovlen; i++)
  {
    xferred += iov->iov_len;
    if (((int)iov->iov_len) < 0 || xferred < 0) /* signed || wrapped */
    {
      errno = EINVAL;
      return -1;
    }
    if (iov->iov_len > 0 && !iov->iov_base)
    {
      #if defined(EFAULT)
      errno = EFAULT;
      #else
      errno = EINVAL;
      #endif
      return -1;
    }
    iov++;
  }
  return 0;
}

/* --------------------------------------------------------------- */

int writev (int fd, const struct iovec *iov, int iovlen)
{
  int i, xferred;
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (socket)
  {
    struct msghdr msg;
    memset( &msg, 0, sizeof(msg));
    msg.msg_iov = (struct iovec *)(*((void **)(&iov)));
    msg.msg_iovlen = iovlen;
    return bsdsys_sendmsg(socket, &msg, 0);
  }
  if (__iov_einval_check(iov, iovlen) != 0)
  {
    return -1;
  }
  xferred = 0;
  for (i = 0; i < iovlen; i++)
  {
    if (iov->iov_len > 0 && iov->iov_base)
    {
      int n = write(fd, iov->iov_base, iov->iov_len);
      if (n == -1) 
      {
        if (xferred == 0)
          return -1;
        break;
      }
      xferred += n;
    }
    iov++;
  }
  return xferred;
}

/* --------------------------------------------------------------- */

int readv (int fd, struct iovec *iov, int iovlen)
{
  int i, xferred;
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (socket)
  {
    struct msghdr msg;
    memset( &msg, 0, sizeof(msg));
    msg.msg_iov = iov;
    msg.msg_iovlen = iovlen;
    return bsdsys_recvmsg(socket, &msg, 0);
  }
  if (__iov_einval_check(iov, iovlen) != 0)
  {
    return -1;
  }
  xferred = 0;
  for (i = 0; i < iovlen; i++)
  {
    if (iov->iov_len > 0 && iov->iov_base)
    {
      int n = read(fd, iov->iov_base, iov->iov_len);
      if (n == -1) 
      {
        if (xferred == 0)
          return -1;
        break;
      }
      xferred += n;
    }
    iov++;
  }
  return xferred;
}

/* =============================================================== */
/*                        Initialization                           */
/* =============================================================== */

#define REDIRECT_LIBCFUNC_WITH_LOCK(__name) \
        redirect_libcfunc( stubdata.isnbuf_##__name , \
                           __name, \
                           __stub_##__name , +1 )
#define RESTORE_LIBCFUNC_WITH_UNLOCK(__name) \
        restore_libcfunc(__name, stubdata.isnbuf_##__name , -1 )

/* --------------------------------------------------------------- */

static void __my_atexit_handler(void)
{
  FSEXT_DPRINTF(("__my_atexit_handler() called\n"));
  if (stubdata.cs_alias)
  {
  /* Don't close sockets/fds. This might be an _exit() */
#   if defined(__DJGPP__) 
    RESTORE_LIBCFUNC_WITH_UNLOCK(write);
#   else
    RESTORE_LIBCFUNC_WITH_UNLOCK(write);
    RESTORE_LIBCFUNC_WITH_UNLOCK(read);
    RESTORE_LIBCFUNC_WITH_UNLOCK(lseek);
    RESTORE_LIBCFUNC_WITH_UNLOCK(close);
#   endif
#   if defined(__WATCOMC__)
    RESTORE_LIBCFUNC_WITH_UNLOCK(__qwrite);
    RESTORE_LIBCFUNC_WITH_UNLOCK(__qread);
#   endif
#   if defined(__TURBOC__)
    RESTORE_LIBCFUNC_WITH_UNLOCK(ioctl);
#   endif
    RESTORE_LIBCFUNC_WITH_UNLOCK(dup2);     /* was redirected for all */
    RESTORE_LIBCFUNC_WITH_UNLOCK(dup);      /* was redirected for all */

    /* --------------------------------- */
    stubdata.cs_alias = 0;
    memset( &stubdata, 0, sizeof(stubdata) );
  }
  return;
}

/* --------------------------------------------------------------- */

int __fsext_setup(void)
{
  int rc = 0;
  __stubdata_init_if_needed();
  if (stubdata.cs_alias == 0)
  {
    rc = -1;
    if (__my_atexit(__my_atexit_handler) == 0)
    {
#if defined(__DPMI32__)     
      stubdata.cs_alias = __my_dpmi_get_csalias();
      if (stubdata.cs_alias == -1)
        stubdata.cs_alias = 0;
      FSEXT_DPRINTF(("cs_alias=0x%04x\n", stubdata.cs_alias ));
      if (stubdata.cs_alias)
#endif
      {
        rc = 0;
        REDIRECT_LIBCFUNC_WITH_LOCK(dup);   /* redirected for all */
        REDIRECT_LIBCFUNC_WITH_LOCK(dup2);  /* redirected for all */
#       if defined(__TURBOC__)
        REDIRECT_LIBCFUNC_WITH_LOCK(ioctl);
#       endif
#       if defined(__WATCOMC__)
        REDIRECT_LIBCFUNC_WITH_LOCK(__qwrite);
        REDIRECT_LIBCFUNC_WITH_LOCK(__qread);
#       endif
#       if defined(__DJGPP__) /* needs crlf fix to write() */
        REDIRECT_LIBCFUNC_WITH_LOCK(write);    
#       else
        REDIRECT_LIBCFUNC_WITH_LOCK(close);
        REDIRECT_LIBCFUNC_WITH_LOCK(lseek);
        REDIRECT_LIBCFUNC_WITH_LOCK(read);
        REDIRECT_LIBCFUNC_WITH_LOCK(write);    
#       endif
      }
    }
  }
  if (rc == -1)
    errno = ENFILE; /* "File table overflow" */
  return rc;
}

