/*
 * BSD Socket<->DOS file description MUX
 *
 * All these functions don't really know anything about the structures
 * handled by the consumer.
 * As far as they are concerned a fd is mapped to "some" opaque object,
 * the pointer to which they manage/associate with a file descriptor.
 *
 * This file consists of functions for allocating/freeing
 * file descriptors and mapping file descriptors to 'struct bsdskt'.
*/

/* +++++ things that need to be done before ANYTHING is included ++++ */
#if defined(__WATCOMC__)
  #define close __close  /* close() and fclose() call __close() */
  #include <io.h>
  #ifdef __cplusplus
  extern "C" {
  #endif
  extern int __IOMode(int fd);
  extern void __SetIOMode(int fd, int mode);
  #ifdef __cplusplus
  }
  #endif
#elif defined(__DJGPP__)
  #include <libc/dosio.h> /* char *__file_handle_modes */
#elif defined(__TURBOC__)
  #define read  ___cooked_read
  #define write ___cooked_write
  #define close ___cooked_close
  #include <io.h>
  #undef read
  #undef write
  #undef close
  #define read  _read
  #define write _write
  #define close _close
  /* extern int _nfile; */
  #ifdef __cplusplus
  extern "C" {
  #endif
  extern int _openfd[];
  #ifdef __cplusplus
  }
  #endif
#endif

#include "target.h"
#include "bsdso.h"
#include "sofdmux.h"

/* =============================================================== */
/*         PART 1 - file descriptor allocation and mapping         */
/* =============================================================== */
/*
 * fd allocation and fd->'struct fdmux_object *' translation functions:
 *
 * All these functions don't really know anything about "sockets"s. 
 * As far as they are concerned a "struct fdmux_object" is just an opaque 
 * object, the pointer to which they manage/associate with a file descriptor.
 *
 * int __sock_fd_alloc(struct fdmux_object FAR *) 
 *     allocates a libc/dos file descriptor for a "socket" (or whatever),
 *     and associates the file descriptor with that pointer.
 *     On success returns a newly allocated file descriptor. 
 *     On error returns -1 and sets errno as appropriate.
 *
 * int __sock_fd_freeall(void)
 *     free all allocated/reserved file descriptors by simply calling
 *     __sock_fd_free for all fds in the association table. 
 *     Nothing more.
 *
 * int __sock_fd_free(int fd)         ++ PUBLIC ++
 *     free's the libc/dos file descriptor if fd was previously associated
 *     with a "socket" (or whatever) using __sock_fd_alloc().
 *
 * struct fdmux_object FAR *__sock_fd_lookup(int fd)   ++ PUBLIC ++
 *     obtain the 'socket' (or whatever) associated with the file 
 *     descriptor 'fd'. Returns NULL if 'fd' is not associated with a socket. 
 *     (errno is unchanged).
 *
 * int __sock_fd_isEBADF(int fd)      ++ PUBLIC ++
 *     returns non-zero if the file descriptor 'fd' is NOT a valid libc/dos 
 *     file descriptor (which may also be a 'socket'). Does not affect errno.
 *
 * int __sock_fd_assoc(int fd, struct fdmux_object FAR *)  ++ PUBLIC ++
 *     assign socket * to an fd.
*/

/* ------------------------------------------------------------------ */

#if defined(__DJGPP__)
  #include <sys/resource.h>
  #include <sys/fsext.h>
  static int __djgpp_fsext_demux(__FSEXT_Fnumber func, int *rv, va_list _args);
  /* __djgpp_fsext_demux is at the end of PART 2. This is the only function
  ** that has to be prototyped separately (it is needed here at fd allocation 
  ** time).
  */
#endif

/* ------------------------------------------------------------------ */

static struct __fd_data
{
  /* DOS permits a maximum of 256 file descriptors per process (0...255) */
  struct fdmux_object FAR *fd_assoc_table[256];
} fdstatics;

#if defined(__TURBOC__)
   #define __DTABLE_SIZE__() ((int)(_nfile)) /* _nfile is a turboc lib thing */
#else
  #define __DTABLE_SIZE__ \
  ((int)(sizeof(fdstatics.fd_assoc_table)/sizeof(fdstatics.fd_assoc_table[0])))
#endif
#define IS_FD_IN_FDSTATICS_RANGE(__fd) \
   ((__fd) >= 0 && (__fd) < __DTABLE_SIZE__)

/* ------------------------------------------------------------------ */

static void __fdstatics_init_if_needed(void)
{
  static int fdstatics_bss_count = -1;

  if (__bss_count != fdstatics_bss_count)
  {
    memset( &fdstatics, 0, sizeof(fdstatics) );
    fdstatics_bss_count = __bss_count;
  }   
  return;
}

/* ------------------------------------------------------------------ */

#if !defined(__DJGPP__)
static int __is_fdtable_growable(void)
{
#if defined(__WATCOMC__)
  int nf = _grow_handles(0);
  if (nf > 0 && IS_FD_IN_FDSTATICS_RANGE(nf) && _grow_handles(nf+5) > nf)
    return 1;
#endif
  return 0;
}
#endif

static int __libc_fd_alloc(void) /* allocate a libc file descriptor */
{
#if defined(__DJGPP__)
  /* same as open("nul") followed by __FSEXT_set_function(fd, _fsext_demux); */
  int fd = __FSEXT_alloc_fd (__djgpp_fsext_demux);
  if (fd == -1)
  {
    struct rlimit r;
    getrlimit (RLIMIT_NOFILE, &r);
    r.rlim_max += 5;
    setrlimit (RLIMIT_NOFILE, &r);
    fd = __FSEXT_alloc_fd (__djgpp_fsext_demux);      
  }
  return fd;
#else
  int fd;
  __fdstatics_init_if_needed();
  /* bypass FILES= limit by dup()ing instead of open()ing */ 
  /* (doesn't mean we can have more than 256 per process of course) */
  for (fd = 0; IS_FD_IN_FDSTATICS_RANGE(fd); fd++)
  {
    if (fdstatics.fd_assoc_table[fd])
    {
      int fd2 = dup(fd);
      if (fd2 != -1 || errno != EMFILE || !__is_fdtable_growable())
        return fd2;
      return dup(fd);
    }
  }
  fd = open("nul", O_RDWR|O_BINARY|O_NOINHERIT );
  if (fd == -1 && errno == EMFILE && __is_fdtable_growable())
    fd = open("nul", O_RDWR|O_BINARY|O_NOINHERIT );
  return fd;
#endif
}

static int __libc_fd_free(int fd) /* free a libc file descriptor */
{
  #if defined(__DJGPP__)
  __FSEXT_set_function(fd, NULL); /* must come before close()! */
  #endif
  return close(fd);
}

/* ------------------------------------------------------------------ */

int __sock_fd_assoc(int fd, struct fdmux_object FAR *socket)
{
  if (IS_FD_IN_FDSTATICS_RANGE(fd)) /* should not be necessary */
  {
    fdstatics.fd_assoc_table[fd] = socket;
    return 0;
  }
  return -1;
}

/* ------------------------------------------------------------------ */

static int __sock_fd_alloc(struct fdmux_object FAR *socket)
{
  int fd = -1;

  __fdstatics_init_if_needed();

  if (!socket)
    errno = EINVAL;
  else
    fd = __libc_fd_alloc();

  if (fd != -1) /* otherwise errno will already have been set */
  {
    if (!IS_FD_IN_FDSTATICS_RANGE(fd)) /* should not be necessary */
    {
      __libc_fd_free(fd);
      #if defined(ENFILE) /* exceeds configured limit */
      errno = ENFILE;
      #else
      errno = EMFILE; /* DOS errno 4 (handle table is full) */
      #endif
      fd = -1;
    }
  }

  if (fd != -1) /* otherwise errno will already be set */  
  {
    fdstatics.fd_assoc_table[fd] = socket;
  }
  return fd;
}

/* ------------------------------------------------------------------ */

static int __sock_fd_reserve(void)
{
  int fd = __libc_fd_alloc();
  if (fd != -1) /* otherwise errno will already have been set */
  {
    __libc_fd_free(fd); /* don't need it anymore */
    if (!IS_FD_IN_FDSTATICS_RANGE(fd)) /* should not be necessary */
    {
      #if defined(ENFILE) /* exceeds configured limit */
      errno = ENFILE;
      #else
      errno = EMFILE; /* DOS errno 4 (handle table is full) */
      #endif
      fd = -1;
    }
  }
  return fd;
}

/* ------------------------------------------------------------------ */

int __sock_fd_free(int fd)
{
  int rc = -1;
  __fdstatics_init_if_needed();
  if (IS_FD_IN_FDSTATICS_RANGE(fd))
  {
    if (fdstatics.fd_assoc_table[fd])
    {    
      __libc_fd_free(fd);
      fdstatics.fd_assoc_table[fd] = (struct fdmux_object FAR *)0;
      rc = 0;
    }
  }
  return rc;
}

/* ------------------------------------------------------------------ */

struct fdmux_object FAR *__sock_fd_lookup(int fd)  /* ++ PUBLIC ++ */
{
  struct fdmux_object FAR *socket = (struct fdmux_object FAR *)0;
  __fdstatics_init_if_needed();
  if (IS_FD_IN_FDSTATICS_RANGE(fd))
    socket = fdstatics.fd_assoc_table[fd];
  return socket;
}   

/* ------------------------------------------------------------------ */

int __sock_fd_isEBADF(int fd)  /* ++ PUBLIC ++ */
{
  int ec = 0; /* assume its not EBADF */
  if (!__sock_fd_lookup(fd)) /* quicker check */
  {
    ec = 6; /* dos err number for EBADF */
    if (IS_FD_IN_FDSTATICS_RANGE(fd)) /* all dos could possibly do too */
    {   
#if defined(__WATCOMC__)
      ec = ((__IOMode(fd) == 0) ? 6 : 0);
#elif defined(PHARLAP_H)
      SWI_REGS reg;
      reg.edx = 0;
      reg.eax = 0x4400;
      reg.ebx = fd;
      _dx_real_int (0x21, &reg);
      if ((reg.flags & 1) == 0)
        reg.eax = 0;
      ec = (reg.eax & 0xffff);
#elif defined(__DPMI32__)
      __dpmi_regs reg;
      reg.d.edx = 0;
      reg.d.eax = 0x4400; 
      reg.d.ebx = fd;
      __dpmi_int (0x21, &reg);
      if ((reg.x.flags & 1)==0)
        reg.d.eax = 0;
      ec = (reg.d.eax & 0xffff);
#else /* real mode targets */
      union REGS reg;
      reg.x.ax = 0x4400;
      reg.x.bx = (short)fd;
      int86 (0x21, &reg, &reg);
      if (!reg.x.cflag)
        reg.x.ax = 0;
      ec = reg.x.ax;
#endif
    } /* if (fd >= 0 && fd <= 255 */
  }
  return (ec == 6); /* is EBADF */
}  


