/*
** netdb routine generation template.
**
** Created June 2000 by Cyrus Patel <cyp@fb14.uni-mainz.de>
**
** The following must be defined before including this file.
**
** define name            example        
** --------------------   -------------
** NDBRGEN_PAYLOAD        struct my_hostent // what functions use internally
** NDBRGEN_RETTYPE        struct hostent // what db functions return (not ptr!)
**  *-*-* the 1st member of NDBRGEN_PAYLOAD must be 'NDBRGEN_RETTYPE' *-*-*
** NDBRGEN_ETCFNCONF      "HOSTS"    // filename option in *.cfg
** NDBRGEN_ENDENTFXNN     endhostent // name of endXXXent function to generate
** NDBRGEN_GETENTFXNN     gethostent // name of getXXXent function to generate
** NDBRGEN_SETENTFXNN     sethostent // name of setXXXent function to generate
** NDBRGEN_SETCANSTAYOPEN 1          // does setXXXent take a 'stayopen' arg?
** NDBRGEN_HAVEALIASES    1          // need db_add_aliases() support
** 
** This include file also generates a function called 
** 'static NDBRGEN_PAYLOAD * enum_db_entries(NDBRGEN_PAYLOAD *prev)' for 
** enumerating the netdb database. To start enumeration, call enum_db_entries
** with NULL. The function will return the next entry or NULL if there are
** no more entries.
**
** If the database had not previously been loaded, enum_db_entries() will 
** first do what is necessary to load it.
**
** The module including this header must have a function called
** 'static NDBRGEN_PAYLOAD * parser_cb(const char *toks[], int numtoks, 
**                                     NDBRGEN_PAYLOAD *resP )'. 
**
** parser_cb() will be called for each (valid) line in the dbfile with a list
** of tokens ("words") on that line, passing a pointer to a preallocated 
** NDBRGEN_PAYLOAD through 'resP'. If the data on the line is valid, parser_cb
** should fill in the passed NDBRGEN_PAYLOAD and return it. If the data is
** invalid, parser_cb() should return NULL. Parsing will continue until
** the whole file has been read, or it runs out of memory.
*/
#if defined(__TURBOC__) /* busted stdlib.h */
#define __MINMAX_DEFINED
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef __DJGPP__
#define memcasecmp memicmp
#define strcasecmp stricmp
#endif

#include "target.h"     /* _my_atexit */
#include "loadcfg.h"    /* _conf_add_confextn */

struct db_entry 
{
  NDBRGEN_PAYLOAD retstruct; /* must come first */
  struct db_entry *next;
};

static struct
{
  #ifdef PATH_MAX
  char db_filename [PATH_MAX+1];
  #elif defined(_MAX_PATH)
  char db_filename [_MAX_PATH+1];
  #else
  #error heeeeelp!
  #endif  
  struct db_entry *entry0;
  struct db_entry *prev_ent;   /* }_ for _getXXXent() */

  int tried_load_once;
} db_statics;

/* ----------------------------------------------------------------------- */

static void __set_etc_fn(const char *value)
{
  if (strlen(value) < sizeof(db_statics.db_filename))
    strcpy( db_statics.db_filename, value );
}

static void __init_statics_if_needed(void)
{
  static int my_bss_count = -1;
  #if defined(__DJGPP__)
  extern int __bss_count;
  #else
  #define __bss_count 0
  #endif
  if (my_bss_count != __bss_count)
  {
    static const struct config_table my_cfg[] = {
       { NDBRGEN_ETCFNCONF, ARG_FUNC, (void FAR *)__set_etc_fn },
       { NULL }
    };
    memset( &db_statics, 0, sizeof(db_statics));

    /* add config table to list of configs to parse */
    _conf_add_confextn( &my_cfg[0], NULL );

    my_bss_count = __bss_count;
  }
  return;
}

#define __CTORNAME(x,y) x ## y
#define _CTORNAME(x,y) __CTORNAME(x,y)
#define CTORNAME(x) _CTORNAME(__ctor_,x)

static void CTORNAME(NDBRGEN_SETENTFXNN) (void)
{ __init_statics_if_needed(); }

#define __SETUP_CONSTRUCTOR_FOR__ CTORNAME(NDBRGEN_SETENTFXNN)
#include "ctordtor.h"

/* ======================================================================== */

static void __unload_dbfile_if_needed(void)
{
  struct db_entry *entry0;
  __init_statics_if_needed();

  entry0 = db_statics.entry0;
  memset( &db_statics, 0, sizeof(db_statics));
  while (entry0)
  {
    struct db_entry *next = entry0->next;
    free(entry0);
    entry0 = next;
  }
  return;
}

/* ----------------------------------------------------------------------- */

/* __tokenize_line() breaks a line from the file, into upto 'numtoks' tokens,
** and stores pointers to these in the tokens[] array. All tokens are '\0'
** terminated. Returns the number of tokens found.
*/
static int __tokenize_line(char *buf, char *tokens[], int numtoks)
{
  int foundtoks = 0;
  while (*buf && *buf != '#' && *buf != '\n')
  {
    char *p;
    while (*buf == ' ' || *buf == '\t' || *buf == ',' || *buf == ';')
      buf++;
    p = buf;
    while (*buf && *buf != '#' && *buf == '\r' || *buf != '\n' && 
           *buf != ' ' && *buf != '\t' && *buf != ',' && *buf != ';')
      buf++;
    if (p != buf)
    {
      char c = *buf;
      *buf++ = '\0';
      tokens[foundtoks++] = p;
      if (foundtoks == numtoks || !c || c == '#' || c == '\r' || c == '\n')
        break;
    }
  }
  return foundtoks;
}

/* ----------------------------------------------------------------------- */

#if 0
static const char *my_getenv(const char *var) /* case sensitive getenv() */
{
  unsigned int i = 0, n = strlen(var);
  while (environ[i])
  {
    const char *p = environ[i];
    if (strlen(p) > n && p[n] == '=' && memcmp(var, p, n)==0)
      return p+n+1;
    i++;
  }
  return (const char *)0;
}
#endif

static NDBRGEN_PAYLOAD *parser_cb(const char *toks[], int numtoks, 
                                  NDBRGEN_PAYLOAD *resP);

static void __load_dbfile_if_needed(void)
{
  __init_statics_if_needed();

  if (!db_statics.tried_load_once)
  {
    FILE *file = (FILE *)0;
    char buf [128];

    _conf_load(); /* load config if not already loaded */

    if (db_statics.db_filename[0]) /* have filename */
    {
      file = fopen(db_statics.db_filename, "rt");
    }
    if (!file)
    {
      const char *p = getenv("OS");
      if (p && strcmp(p, "Windows_NT")==0)
        p = getenv("SystemRoot");
      if (!p)
        p = getenv ("windir");
      if (p)
      {
        /* On Win95/98 use `hosts = $(WINDIR)\system\hosts' */
        /* On WinNT/2K use `hosts = $(WINDIR)\system32\drivers\etc\hosts' */
        if ((strlen(p) + sizeof("\\system\\xxxxxxxx")) < sizeof(buf))
        {
          strcat( strcpy( buf, p ),"\\system\\" );
          file = fopen( strcat(buf, NDBRGEN_ETCFNCONF), "rt");
        }
        if (!file && (strlen(p) +
             sizeof("\\system32\\drivers\\etc\\xxxxxxxx")) < sizeof(buf))
        {
          strcat( strcpy( buf, p ), "\\system32\\drivers\\etc\\" );
          file = fopen( strcat(buf, NDBRGEN_ETCFNCONF), "rt");
        }
      }
    }

    if (file) /* got a filename */
    {
      struct db_entry *dbe = (struct db_entry *)0;
      struct db_entry *tail = (struct db_entry *)0;
      while (fgets (buf, sizeof(buf)-1, file))
      {
        char *tokens[16]; 
        int numtoks;

        if (!strchr(buf, '\n'))
        {
          int c = 1;
          while (c != '\n' && c != EOF)
            c = fgetc(file);
        }

        numtoks = (sizeof(tokens)/sizeof(tokens[0]))-1;
        numtoks = __tokenize_line(buf, &tokens[0], numtoks);
        tokens[numtoks] = (char *)0;

        if (numtoks)
        {
          if (!dbe)
          {
            dbe = (struct db_entry *)calloc(sizeof(*dbe),1);
            if (!dbe)
              break;
          }            
          if (parser_cb( ((const char **)(&tokens[0])), numtoks, 
                         &(dbe->retstruct) )) 
          {
            /* line was accepted */
            dbe->next = (struct db_entry *)0;
            if (tail)
              tail->next = dbe;
            else
            {
              db_statics.entry0 = dbe;
              _my_atexit (__unload_dbfile_if_needed); 
            }
            tail = dbe;
            dbe = dbe->next;
          } 
        } /* if (numtoks) */
      } /* while (fgets (buf, sizeof(buf)-1, file)) */
      if (dbe)
        free(dbe);
      fclose(file); 
    } /* if (file) */

    db_statics.tried_load_once = 1;
  } /* if (!db_statics.tried_load_once) */

  return;
}

/* ======================================================================= */
/*                getXXXent(), setXXXent(), endXXXent()                    */
/* ======================================================================= */

void NDBRGEN_ENDENTFXNN (void)
{
  __init_statics_if_needed();
  db_statics.prev_ent = (struct db_entry *)0;
}

#if NDBRGEN_SETCANSTAYOPEN
void NDBRGEN_SETENTFXNN (int stayopen) {
  stayopen = stayopen;  /* unused */
#else
void NDBRGEN_SETENTFXNN (void) {
#endif
  NDBRGEN_ENDENTFXNN ();
}

NDBRGEN_RETTYPE * NDBRGEN_GETENTFXNN (void)
{
  __init_statics_if_needed();
  __load_dbfile_if_needed();

  if (!db_statics.prev_ent) /* top of file */
  {
    db_statics.prev_ent = db_statics.entry0;
    return (NDBRGEN_RETTYPE *)&(db_statics.prev_ent->retstruct);
  }
  else if (db_statics.prev_ent->next)
  {
    db_statics.prev_ent = db_statics.prev_ent->next;
    return (NDBRGEN_RETTYPE *)&(db_statics.prev_ent->retstruct);
  }
  /* hit eof */
  return (NDBRGEN_RETTYPE *)0; 
}

/* ======================================================================= */
/*                the only "exported" functions. simple, eh?               */
/* ======================================================================= */

#if NDBRGEN_HAVEALIASES
/* tokenizer support */
static int db_add_aliases(const char *tokens[], int numtoks, 
                          char *aliasbuf, unsigned int aliasbuflen,
                          char **aliaslist, unsigned int aliaslistsz)
{
  unsigned int aliascount = 0, len = 0;
  int tokpos = 0; 
  while (tokpos < numtoks && aliascount < (aliaslistsz-1) )
  {
    const char *src = tokens[tokpos++];
    unsigned int srclen;

    if (!src) /* hmm! */
      break;

    srclen = strlen(src);
    if ((len + srclen + 2) < aliasbuflen)
    {
      strcpy( &(aliasbuf[len]), src );
      aliaslist[aliascount++] = &(aliasbuf[len]); 
      len += srclen + 1;
    }
  }
  aliaslist[aliascount] = (char *)0;
  return aliascount;
}
#endif

static NDBRGEN_PAYLOAD * enum_db_entries(NDBRGEN_PAYLOAD * prev)
{
  __init_statics_if_needed();
  __load_dbfile_if_needed();
  if (!prev)
  {
    if (!db_statics.entry0)
      return (NDBRGEN_PAYLOAD *)0;
    return (NDBRGEN_PAYLOAD *)db_statics.entry0;
  }
  return (NDBRGEN_PAYLOAD *)(((struct db_entry *)prev)->next);
}


/* ======================================================================= */

