/*
** gethostbyname() and gethostbyaddr()
** 
** Created June 2000 by Cyrus Patel <cyp@fb14.uni-mainz.de>
*/
#include <netinet/in.h> /* struct in_addr */
#include <netdb.h>      /* h_errno, gethostbyxxx(), gethostname() */
#include <arpa/inet.h>  /* inet_addr */
#include <sys/socket.h> /* AF_INET */

struct my_hostent
{
  struct hostent he;
  char   namebuf[128];
  struct in_addr addrbuf[1];
  char   *addrlist[2];
  int    addrcount;
  char   aliasbuf[128];
  char   *aliaslist[5];
  int    aliascount;
};

#define NDBRGEN_PAYLOAD        struct my_hostent  /* what we deal with */
#define NDBRGEN_RETTYPE        struct hostent     /* what we return (not ptr)*/
#define NDBRGEN_ETCFNCONF      "HOSTS"     /* entry in wattcp.cfg */
#define NDBRGEN_ENDENTFXNN     endhostent  /* what to generate for endXXXent() */
#define NDBRGEN_GETENTFXNN     gethostent  /* what to generate for getXXXent() */
#define NDBRGEN_SETENTFXNN     sethostent  /* what to generate for setXXXent() */
#define NDBRGEN_SETCANSTAYOPEN 1  /* does setXXXent take a 'stayopen' arg? */
#define NDBRGEN_HAVEALIASES    1  /* yes, we need alias support */

#include "dbentgen.h" /* generate all the necessary db access skeleton code */ 
#include "libtso.h"   /* _dns_resbyname(), _dns_resbyaddr() */

/* ======================================================================= */

/* This is our entry parser called by the auto-generated code to 
** parse a line in "/etc/hosts". 
**
** On fatal error (ie out of memory), return -1 (the parsing will stop). 
** On success return zero and set *resPP = the newly generated struct
** (or NULL if we didn't generate one because of an invalid token).
*/
static NDBRGEN_PAYLOAD *parser_cb(const char *tokens[], int numtoks, 
                                  NDBRGEN_PAYLOAD *resP)
{
  NDBRGEN_PAYLOAD *e = (NDBRGEN_PAYLOAD *)0;
  if (numtoks >= 2)
  {
    if (strlen(tokens[1]) < (sizeof(e->namebuf)-1))
    {
      struct in_addr addr;
      addr.s_addr = inet_addr(tokens[0]);
      if (addr.s_addr != 0 && addr.s_addr != 0xfffffffful)
      {
        e = resP;
        e->addrbuf[0].s_addr = addr.s_addr;
        strcpy (e->namebuf, tokens[1]);
        e->addrlist[0]    = (char *)&(e->addrbuf[0]);
        e->addrlist[1]    = (char *)0;
        e->addrcount      = 1;
        e->aliaslist[0]   = (char *)0;    
        e->aliascount     = db_add_aliases(&tokens[2], numtoks-2, 
                              e->aliasbuf, sizeof(e->aliasbuf),
                              &(e->aliaslist[0]), 
                              sizeof(e->aliaslist)/sizeof(e->aliaslist[0]) );
        e->he.h_name      = &(e->namebuf[0]);
        e->he.h_aliases   = &(e->aliaslist[0]);
        e->he.h_addrtype  = AF_INET;
        e->he.h_length    = sizeof(struct in_addr);
        e->he.h_addr_list = &(e->addrlist[0]);

      } /* ip addr is valid */
    } /* hostname fits in buffer */
  } /* if (numtoks >= 2) */
  return e;
}

/* ======================================================================= */
/*  internal functions to support one set of code for both _r and non _r   */
/* ======================================================================= */

struct r_args
{                         /* when a *_r function was called, the members */
  struct hostent *result; /* are the same the arguments of that _r function, */
  char *buffer;           /* otherwise, for non-reentrant functions, they */
  int buflen;             /* point to static storage (or TLS space) */
  int *h_errnop;
};

/* ---------------------------------------------------------------------- */

static struct r_args *__make_r_buffer(int make_from_tls, 
                                      struct r_args *r_buf, 
                                      struct hostent *result, 
                                      char *buffer, int buflen, 
                                      int *h_errnop )
{
  if (make_from_tls)
  {
    static struct hostent tls_result;
#define __MAX_ADDR_CACHE 8 /* max number of addresses to store */
    static char tls_buffer[ 256+1+ /* hostname */
                 ((__MAX_ADDR_CACHE+1)*sizeof(char *))+ /* addrlist */
                 ((__MAX_ADDR_CACHE)*sizeof(struct in_addr))+ /* addr */
                 ((7)*sizeof(char *))+ /* alias list */
                 256+1+ /* alias buffer */
                 sizeof(char *) /* padding */];
#undef __MAX_ADDR_CACHE
    result  = &tls_result; 
    buffer  = &tls_buffer[0];
    buflen  = sizeof(tls_buffer);
    h_errnop = &(h_errno);
  }
  if (!buffer || !result)
    buflen = 0;
  else if (buflen >= ((int)sizeof(char *))) 
  {
    /* make sure the buffer is aligned */
    unsigned long aoff = ((unsigned long)buffer);
    if ( (aoff & (sizeof(char *)-1)) != 0)
    {
      aoff = (sizeof(char *)-((size_t)(aoff & (sizeof(char *)-1))));
      buffer += (size_t)aoff;
      buflen -= (size_t)aoff;
    }
  }
  if (!buffer || buflen < ((int)sizeof(char *)) || !result) 
  {
    errno = EINVAL;
    if (h_errnop)
      *h_errnop = -1;
    return (struct r_args *)0;
  }    
  r_buf->result = result;    
  r_buf->buffer = buffer;
  r_buf->buflen = buflen;
  r_buf->h_errnop = h_errnop;
  return r_buf;
}

/* ---------------------------------------------------------------------- */

/* apply h_errno and (optionally) an errno */
static void __insert_h_errno( struct r_args *_r_buffer,
                              int my_h_errno, int my_errno)
{
  if (_r_buffer)
  {
    if (_r_buffer->h_errnop)
      *(_r_buffer->h_errnop) = my_h_errno;
  }      
  if (my_h_errno == -1) /* NETDB_INTERNAL */
  {
    if (my_errno > 0) /* otherwise its already set */
      errno = my_errno;
  }    
  return;
}

/* ---------------------------------------------------------------------- */

/* split/parse/stuff the results from a lookup into the r_args space
*/    
static struct hostent *__insert_result( struct r_args * _r_buffer,
                                        const char *hostname,
                                        const struct in_addr *addr_list, 
                                        unsigned int addr_count,
                                        const char **alias_list,
                                        unsigned int alias_count,
                                        int gen_errno )
{
  struct hostent *hp = (struct hostent *)0;
  if (!addr_count)
  {
    if (_r_buffer) /* should always be so (done in __gethostbyXXYY__) */
    {
      if (_r_buffer->h_errnop)
        *(_r_buffer->h_errnop) = gen_errno;
    }  
  }
  else if (_r_buffer) /* should always be so (done in __gethostbyXXYY__) */
  {
    static char *static_null_pointer = (char *)0; /* dummy */
    int min_needed = strlen(hostname)+1+sizeof(char *)+  /* h_name */
                     sizeof(char **)+(sizeof(char *)*2)+ /* h_addr_list */
                     sizeof(addr_list[0]); /* h_addr_list[0] */

    if (!_r_buffer->buffer || !_r_buffer->result)
      _r_buffer->buflen = 0;
    else if (_r_buffer->buflen > (int)(sizeof(char *)))
    {  
      /* make sure the buffer is aligned */
      unsigned long aoff = ((unsigned long)_r_buffer->buffer);
      if ((aoff & (sizeof(char *)-1)) != 0)
      {
        aoff = (sizeof(char *)-((size_t)(aoff & (sizeof(char *)-1))));
        _r_buffer->buffer += (size_t)aoff;
        _r_buffer->buflen -= (size_t)aoff;
      }        
    }
/* printf("buflen=%d, min_needed=%d\n", _r_buffer->buflen, min_needed ); */
    
    if ((min_needed >= _r_buffer->buflen) || !_r_buffer->result)
    {
#if defined(ERANGE)
      errno = ERANGE;
#endif
      if (_r_buffer->h_errnop)
        *(_r_buffer->h_errnop) = -1; /* NETDB_INTERNAL */
    }
    else
    {
      char *datap;
      unsigned int max_addrs = 0;
      unsigned int max_aliases = 0;
      unsigned int i; int addit_needed;
      int namelen, buflen = _r_buffer->buflen;

      for (i = 0; min_needed < buflen && i < addr_count; i++)
      {
        addit_needed = sizeof(char *)+sizeof(addr_list[0]);
        if (i == 0)
          addit_needed += sizeof(char **)+sizeof(char *);
        min_needed += addit_needed;
        if (min_needed < buflen)
          max_addrs++;
      }        
/* printf("max_addrs=%u\n", max_addrs); */
      for (i = 0; min_needed < buflen && i < alias_count; i++)
      {
        addit_needed = sizeof(char *)+strlen(alias_list[i])+1;
        if (i == 0)
          addit_needed += sizeof(char **)+sizeof(char *);
        min_needed += addit_needed;
        if (min_needed < buflen)
          max_aliases++;          
      }         
/* printf("max_aliases=%u\n", max_aliases); */

      hp = _r_buffer->result;
      datap = _r_buffer->buffer; /* already aligned */
      
      static_null_pointer = (char *)0;
      memset( hp, 0, sizeof(struct hostent));
      hp->h_addrtype = AF_INET;
      hp->h_length = sizeof(struct in_addr);
      hp->h_name = "";
      hp->h_aliases = &static_null_pointer;
      hp->h_addr_list = &static_null_pointer;
      
      if (hostname[0])
      {
        hp->h_name = datap;
        namelen = strlen( strcpy( datap, hostname ) );
        if (datap[namelen-1] == '.') /* canonical */
          datap[--namelen] = '\0';
        datap += 1 + namelen;
      }
      if (max_addrs)
      {
        datap += (size_t)(sizeof(char *)-
                    ((datap-(_r_buffer->buffer))&(sizeof(char *)-1)));
        hp->h_addr_list = (char **)datap;
        datap += (sizeof(char *) * (max_addrs+1));
        for (i = 0; i < max_addrs; i++)
        {
          hp->h_addr_list[i] = datap;
          memcpy( datap, &addr_list[i], sizeof(addr_list[i]) );
          datap += sizeof(addr_list[i]);
        }
        hp->h_addr_list[max_addrs] = (char *)0;
        hp->h_addr_list = &(hp->h_addr_list[0]);
      }
      if (max_aliases)
      {
        datap += (size_t)(sizeof(char *)-
                   ((datap-(_r_buffer->buffer))&(sizeof(char *)-1)));
        hp->h_aliases = (char **)datap;
        datap += ((max_aliases+1)*sizeof(char *));
        for (i = 0; i < max_aliases; i++)
        {
          hp->h_aliases[i] = datap;
          strcpy( datap, alias_list[i] );
          namelen = strlen( datap );
          if (namelen && datap[namelen-1] == '.') /* canonical */
            datap[--namelen] = '\0';
          datap += 1 + namelen;
        }
        hp->h_aliases[max_aliases] = (char *)0;
        hp->h_aliases = &(hp->h_aliases[0]);
      }
      if (_r_buffer->h_errnop)
        *(_r_buffer->h_errnop) = 0; /* NETDB_SUCCESS */
    }
  }
/* printf("return hp=%p\n", hp ); */
  return hp;
}

/* ======================================================================= */
/*                         get data from /etc/hosts                        */
/* ======================================================================= */

static int __i_etc_gethostbyname2(struct r_args *_r_buffer,
                                  const char *name, int af)
{
  struct my_hostent *e;

  if (af != AF_INET)
  {
    __insert_h_errno( _r_buffer, -1 /*NETDB_INTERNAL*/,EAFNOSUPPORT);
    return -1;
  }

  /* note that _etc_gethostname() fails if 'name' is an IP address - 
  ** _etc_gethostbyname() *only* searches the /etc/hosts file.
  */

  /* begin enumeration by passing NULL to enum_db_entries */
  e = enum_db_entries((struct my_hostent *)0);
  while (e)
  {
    unsigned int i = 0;
    if (strcasecmp( name, e->he.h_name) == 0)
      break;
    while (e->he.h_aliases[i])
    {
      if (strcasecmp(name, e->he.h_aliases[i]) == 0)
        break;
      i++;       
    }
    if (e->he.h_aliases[i])
      break;
    e = enum_db_entries(e); /* get next */
  }

  if (e)
  {
    __insert_result( _r_buffer, e->he.h_name, &(e->addrbuf[0]), e->addrcount,
                     (const char **)&(e->he.h_aliases[0]), e->aliascount, 0 );
    return 1;
  }

  __insert_h_errno( _r_buffer, HOST_NOT_FOUND, 0 );
  return 0;
}

/* --------------------------------------------------------------------- */

static int __i_etc_gethostbyaddr(struct r_args *_r_buffer,
                                 const char *addr, size_t len, int af)
{
  struct my_hostent *e;

  if (af != AF_INET)
  {
    __insert_h_errno( _r_buffer, -1 /*NETDB_INTERNAL*/,EAFNOSUPPORT);
    return -1;
  }

  if (len != sizeof(struct in_addr))
  {
    __insert_h_errno( _r_buffer, -1 /*NETDB_INTERNAL*/, EINVAL);
    return -1;
  }

  /* we'd never find these, so don't bother looking */
  if (((struct in_addr *)addr)->s_addr == 0 || 
      ((struct in_addr *)addr)->s_addr == 0xfffffffful)
  {
    __insert_h_errno( _r_buffer, HOST_NOT_FOUND, 0);
    return -1;
  }

  /* begin enumeration by passing NULL to enum_db_entries */
  e = enum_db_entries((struct my_hostent *)0);
  while (e)
  {
    unsigned int i = 0;
    while (e->he.h_addr_list[i])
    {
      if (memcmp(addr, e->he.h_addr_list[i], len) == 0)
        break;
      i++;       
    }
    if (e->he.h_addr_list[i])
      break;
    e = enum_db_entries(e); /* get next */
  }

  if (e)
  {
    __insert_result( _r_buffer, e->he.h_name, &(e->addrbuf[0]), e->addrcount,
                     (const char **)&(e->he.h_aliases[0]), e->aliascount, 0 );
    return 1;
  }

  __insert_h_errno( _r_buffer, HOST_NOT_FOUND, 0 );
  return 0;
}

/* ======================================================================= */
/*                         get data from DNS                               */
/* ======================================================================= */

static int __cvt_res2netdb_err(int err)
{
  switch (err)
  {
    case  0: /* not found (authoritative) */
             return HOST_NOT_FOUND;
    case -1: /* format error */ 
             return NO_RECOVERY;
    case -2: /* server failure or no timely response from server */
             return NO_DATA;
    case -3: /* request cancelled (EINTR) */
             return TRY_AGAIN;
    case -4: /* network error */
             return TRY_AGAIN;
    case -5: /* request refused */
             return NO_RECOVERY;
    default:
             err = 0;
  }
  return NO_RECOVERY;
}

/* -------------------------------------------------------------------- */

#undef  INADDR_LOOPBACK
#define INADDR_LOOPBACK  (0x7F000001ul)

static int __i_dns_gethostbyname2(struct r_args *_r_buffer,
                                 const char *name, int af)
{
  char scratch[128];
  struct in_addr addr[8];
  int numaddrs = 0;
  int err = HOST_NOT_FOUND;

  if (af != AF_INET)
  {
    __insert_h_errno( _r_buffer, -1 /*NETDB_INTERNAL*/,EAFNOSUPPORT);
    return -1;
  }

  if (strlen(name) >= (sizeof(scratch)-1))
  {
    __insert_h_errno( _r_buffer, -1 /*NETDB_INTERNAL*/, ERANGE);
    return -1;
  }

  addr[0].s_addr = inet_addr(name);

  if (addr[0].s_addr != 0xfffffffful) /* its already an address */
  {
    numaddrs = 1; 
  }
  else if (strcasecmp(name, "lb")==0 ||
     (strlen(name) >= 9 && (name[9] == '\0' || name[9] == '.') &&
     memcasecmp(name,"localhost",9)==0))
  {
    addr[0].s_addr = INADDR_LOOPBACK;
    numaddrs = 1;
  }
  else 
  {
    if (gethostname(scratch, sizeof(scratch))==0)
    {
      scratch[sizeof(scratch)-1] = '\0';
      if (strcasecmp(scratch, name) == 0)
      {
        addr[0].s_addr = gethostid();
        numaddrs = 1;
        name = scratch;
      }
    }
  }

  if (numaddrs == 0)
  {
    numaddrs = _dns_resbyname( name, (DWORD *)&addr[0], 
                                (sizeof(addr)/sizeof(addr[0])) );
    if (numaddrs > 0)
    {
      int i;
      for (i = 0; i < numaddrs; i++)
        addr[i].s_addr = htonl(addr[i].s_addr);
    }
    else if (numaddrs < 0)
    {
      err = __cvt_res2netdb_err(numaddrs);
      numaddrs = 0;
    }
  }

  if (numaddrs != 0) /* successfully resolved */
  {
    const char *alias = "";
    __insert_result( _r_buffer, name, &addr[0], numaddrs, &alias, 0, 0);
    return 1;
  }

  __insert_h_errno( _r_buffer, err, 0);
  return 0;
}

/* --------------------------------------------------------------------- */

static int __i_dns_gethostbyaddr(struct r_args *_r_buffer,
                                 const char *addrP, size_t len, int af)
{
  char scratch[128];
  struct in_addr addr;
  int err = HOST_NOT_FOUND;

  if (af != AF_INET)
  {
    __insert_h_errno( _r_buffer, -1 /*NETDB_INTERNAL*/,EAFNOSUPPORT);
    return -1;
  }

  if (len != sizeof(struct in_addr))
  {
    __insert_h_errno( _r_buffer, -1 /*NETDB_INTERNAL*/,EINVAL);
    return -1;
  }

  addr.s_addr = ((struct in_addr *)addrP)->s_addr;
  scratch[0] = '\0';

  if (addr.s_addr == 0 || addr.s_addr == 0xfffffffful)
  {                               /* never succeeds - discard now */
    __insert_h_errno( _r_buffer, HOST_NOT_FOUND, 0);
    return -1;
  }

  if (scratch[0] == '\0' && addr.s_addr == INADDR_LOOPBACK)
  {
    strcpy(scratch, "localhost");
  }

  if (scratch[0] == '\0' && addr.s_addr == gethostid())
  {
    if (gethostname(scratch, sizeof(scratch)) != 0)
      scratch[0] = '\0';
    else
      scratch[sizeof(scratch)-1] = '\0';
  }

  if (scratch[0] == '\0')
  {
    int err = _dns_resbyaddr( ntohl(addr.s_addr), scratch, sizeof(scratch));
    if (err <= 0)
    {
      err = __cvt_res2netdb_err(err);
      scratch[0] = '\0';
    }
  }

  if (scratch[0]) /* successfully resolved */
  {
    const char *alias = "";
    __insert_result( _r_buffer, scratch, &addr, 1, &alias, 0, 0);
    return 1;
  }  

  __insert_h_errno( _r_buffer, err, 0);
  return 0;
}

/* ===================================================================== */
/*                  /etc/hosts or dns lookup selectors                   */
/* ===================================================================== */

static struct hostent * _XXX_gethostbyname2(int do_dns, 
                                            const char *name, int af)
{
  struct r_args r_buf;
  if (!__make_r_buffer( 1, &r_buf, 0, 0, 0, 0))
    return (struct hostent *)0; /* errno/h_errno is already set */
  if (do_dns)
  {
    if (__i_dns_gethostbyname2( &r_buf, name, af ) > 0)
      return r_buf.result;    
  } else if (__i_etc_gethostbyname2( &r_buf, name, af ) > 0)
      return r_buf.result;
  return (struct hostent *)0; /* errno/h_errno is already set */  
}

static struct hostent * _XXX_gethostbyname2_r(int do_dns, 
                                              const char *name, int af,
                                              struct hostent *result, 
                                              char *buffer, int buflen, 
                                              int *h_errnop )
{
  struct r_args r_buf;
  if (!__make_r_buffer(0, &r_buf, result, buffer, buflen, h_errnop))
    return (struct hostent *)0; /* errno/h_errno is already set */
  if (do_dns)
  {
    if (__i_dns_gethostbyname2( &r_buf, name, af ) > 0)
      return r_buf.result;    
  } else if (__i_etc_gethostbyname2( &r_buf, name, af ) > 0)
      return r_buf.result;
  return (struct hostent *)0; /* errno/h_errno is already set */  
}

static struct hostent * _XXX_gethostbyaddr_r(int do_dns,
                                         const char *addr, size_t len, int af,
                                         struct hostent *result, 
                                         char *buffer, int buflen, 
                                         int *h_errnop )
{
  struct r_args r_buf;
  if (!__make_r_buffer(0, &r_buf, result, buffer, buflen, h_errnop))
    return (struct hostent *)0; /* errno/h_errno is already set */
  if (do_dns)
  {
    if (__i_dns_gethostbyaddr( &r_buf, addr, len, af ) > 0)
      return r_buf.result;
  } else if (__i_etc_gethostbyaddr( &r_buf, addr, len, af ) > 0)
    return r_buf.result;
  return (struct hostent *)0; /* errno/h_errno is already set */  
}

static struct hostent * _XXX_gethostbyaddr(int do_dns, 
                                         const char *addr, size_t len, int af)
{
  struct r_args r_buf;
  if (!__make_r_buffer( 1, &r_buf, 0, 0, 0, 0 ))
    return (struct hostent *)0; /* errno/h_errno is already set */
  if (do_dns)
  {
    if (__i_dns_gethostbyaddr( &r_buf, addr, len, af ) > 0)
      return r_buf.result;
  } else if (__i_etc_gethostbyaddr( &r_buf, addr, len, af ) > 0)
    return r_buf.result;
  return (struct hostent *)0; /* errno/h_errno is already set */  
}

/* ===================================================================== */
/*               exposed Name service switch functions                   */
/* ===================================================================== */

struct hostent * _etc_gethostbyname2_r(const char *name, int af,
                                      struct hostent *result, 
                                      char *buffer, int buflen, 
                                      int *h_errnop )
{
  return _XXX_gethostbyname2_r(0, name, af, result, buffer, buflen, h_errnop);
}

struct hostent * _dns_gethostbyname2_r(const char *name, int af,
                                      struct hostent *result, 
                                      char *buffer, int buflen, 
                                      int *h_errnop )
{
  return _XXX_gethostbyname2_r(1, name, af, result, buffer, buflen, h_errnop);
}


struct hostent * _etc_gethostbyname_r(const char *name,
                                      struct hostent *result, 
                                      char *buffer, int buflen, 
                                      int *h_errnop )
{
  return _XXX_gethostbyname2_r(0,name,AF_INET,result,buffer,buflen,h_errnop);
}

struct hostent * _dns_gethostbyname_r(const char *name,
                                      struct hostent *result, 
                                      char *buffer, int buflen, 
                                      int *h_errnop )
{
  return _XXX_gethostbyname2_r(1,name,AF_INET,result,buffer,buflen,h_errnop);
}

struct hostent * _etc_gethostbyname2(const char *name, int af)
{
  return _XXX_gethostbyname2(0,name,af);
}

struct hostent * _dns_gethostbyname2(const char *name, int af)
{
  return _XXX_gethostbyname2(1,name,af);
}

struct hostent * _etc_gethostbyname(const char *name)
{
  return _XXX_gethostbyname2(0,name,AF_INET);
}

struct hostent * _dns_gethostbyname(const char *name)
{
  return _XXX_gethostbyname2(1,name,AF_INET);
}

struct hostent * _etc_gethostbyaddr_r(const char *addr, size_t len, int af,
                                      struct hostent *result, 
                                      char *buffer, int buflen, 
                                      int *h_errnop )
{
  return _XXX_gethostbyaddr_r(0,addr,len,af,result,buffer,buflen,h_errnop);
}

struct hostent * _dns_gethostbyaddr_r(const char *addr, size_t len, int af,
                                      struct hostent *result, 
                                      char *buffer, int buflen, 
                                      int *h_errnop )
{
  return _XXX_gethostbyaddr_r(1,addr,len,af,result,buffer,buflen,h_errnop);
}

struct hostent * _etc_gethostbyaddr(const char *addr, size_t len, int af)
{
  return _XXX_gethostbyaddr(0, addr, len, af);
}

struct hostent * _dns_gethostbyaddr(const char *addr, size_t len, int af)
{
  return _XXX_gethostbyaddr(1, addr, len, af);
}

/* ===================================================================== */
/*                   traditional name lookup functions                   */
/* ===================================================================== */

struct hostent * gethostbyname2 (const char *name, int af)
{
  struct hostent *hp = _etc_gethostbyname2(name, af);
  if (!hp && h_errno == HOST_NOT_FOUND)
    hp = _dns_gethostbyname2(name, af);
  return hp;
}  

struct hostent * gethostbyname2_r (const char *name, int af,
                                      struct hostent *result, 
                                      char *buffer, int buflen, 
                                      int *h_errnop )
{
  struct hostent *hp; int dummy = 0; 
  if (!h_errnop) h_errnop = &dummy;
  hp = _etc_gethostbyname2_r (name, af, result, buffer, buflen, h_errnop);
  if (!hp && *h_errnop == HOST_NOT_FOUND)
    hp = _dns_gethostbyname2_r (name, af, result, buffer, buflen, h_errnop);
  return hp;
}  


struct hostent * gethostbyname (const char *name)
{
  return gethostbyname2 (name, AF_INET);
}

struct hostent * gethostbyname_r (const char *name,
                                  struct hostent *result, 
                                  char *buffer, int buflen, 
                                  int *h_errnop )
{
  return gethostbyname2_r (name, AF_INET, result, buffer, buflen, h_errnop);
}

struct hostent * gethostbyaddr (const char *addr, int len, int af)
{                                               /* len should be size_t */
  struct hostent *hp = _etc_gethostbyaddr(addr, len, af);
  if (!hp && h_errno == HOST_NOT_FOUND)
    hp = _dns_gethostbyaddr(addr, len, af);
  return hp;
}

struct hostent * gethostbyaddr_r (const char *addr, size_t len, int af,
                                  struct hostent *result, 
                                  char *buffer, int buflen, 
                                  int *h_errnop )
{
  struct hostent *hp; int dummy = 0; 
  if (!h_errnop) h_errnop = &dummy;
  hp = _etc_gethostbyaddr_r (addr, len, af, result, buffer, buflen, h_errnop);
  if (!hp && *h_errnop == HOST_NOT_FOUND)
    hp = _dns_gethostbyaddr_r(addr, len, af, result, buffer, buflen, h_errnop);
  return hp;
}

/* ====================================================================== */

#if defined(TEST_PROG)
static void dump_ent(const char *label, const struct hostent *e)
{
  if (!e)
    printf("%s: no data!\n", label);
  else
  {
    int i;
    printf ("%s: h_name='%s' h_addrtype=%d, h_length=%d\n", 
              label, e->h_name, e->h_addrtype, e->h_length );
    for (i = 0; e->h_aliases && e->h_aliases[i]; i++)
      printf("        h_aliases[%d]='%s'\n", i, e->h_aliases[i] );
    for (i = 0; e->h_addr_list && e->h_addr_list[i]; i++)
      printf("        h_addr_list[%d]=%s\n", i, 
                      inet_ntoa( *((struct in_addr *)(e->h_addr_list[i])) ) );
  }
  return;
}

int main (int argc, char *argv[])
{
  if (argc > 1)
  {
    struct in_addr addr;
    if (strcmp(argv[1],"all")==0)
      argc = 1; /* fall through */
    else if (inet_aton(argv[1], &addr))
      dump_ent("byaddr", gethostbyaddr( (char *)&addr, sizeof(addr), AF_INET ));
    else
      dump_ent("byname", gethostbyname( argv[1] ));
  }
  if (argc == 1)
  {
    struct hostent *e;
    while ((e = gethostent()) != NULL)
      dump_ent("getent", e);
  }
  return (0);
}
#endif /* TEST_PROG */
