/*
** getprotobyname() and getprotobynumber()
** 
** Created June 2000 by Cyrus Patel <cyp@fb14.uni-mainz.de>
*/
#include <netdb.h> /* h_errno */

struct my_protoent {
    struct protoent pe;
    char namebuf[32];
    char aliasbuf[64];
    char *aliaslist[5];
};

#define NDBRGEN_PAYLOAD        struct my_protoent  /* what we deal with internally*/
#define NDBRGEN_RETTYPE        struct protoent  /* what we return (not ptr!) */
#define NDBRGEN_ETCFNCONF      "PROTOCOLS"      /* entry in wattcp.cfg */
#define NDBRGEN_ENDENTFXNN     endprotoent     /* what to generate for endXXXent() */
#define NDBRGEN_GETENTFXNN     getprotoent     /* what to generate for getXXXent() */
#define NDBRGEN_SETENTFXNN     setprotoent     /* what to generate for setXXXent() */
#define NDBRGEN_SETCANSTAYOPEN 1  /* does setXXXent take a 'stayopen' arg? */
#define NDBRGEN_HAVEALIASES    1  /* yes, we need alias support */

#include "dbentgen.h" /* generate all the necessary db access skeleton code */ 

/* ======================================================================= */

/* provide a builtin table for convenience. Note that getprotoent() will
** never retrieve a protoent from here (correctly so).
*/
static struct protoent * scan_builtins(const char *name, int proto)
{
  static const char *builtins[] = { "\0""ip" ,    
                                    "\1""icmp", 
                                    "\2""igmp",
                                    "\6""tcp",    
                                    "\8""egp",  
                                    "\9""igp", 
                                    "\16""chaos", 
                                    "\17""udp", 
                                    "\41""ipv6",
                                    "\43""ipv6-route", 
                                    "\46""ipv6-frag",
                                    "\58""ipv6-icmp",
                                    "\59""ipv6-nonxt",
                                    "\60""ipv6-opts" };
  unsigned i = 0;
  for (i = 0; i < (sizeof(builtins)/sizeof(builtins[0]));i++)
  {
    if ( (name && strcasecmp(name, (builtins[i]+1))==0) 
       || (!name && *builtins[i] == ((char)proto)) )
    {
      static char *aliases[1];
      static struct protoent pe;
      union { const char *ccp; char *cp; } ccpcp;
      aliases[0] = (char *)0;
      ccpcp.ccp = builtins[i]+1;
      pe.p_name = ccpcp.cp;
      pe.p_proto = 0xff & ((int)(*builtins[i]));
      pe.p_aliases = &aliases[0];
      return &pe;
    }
  }
  return (struct protoent *)0;
}

NDBRGEN_RETTYPE * getprotobyname (const char *name)
{
  if (name)
  {
    struct protoent *pe;
    /* begin enumeration by passing NULL to enum_db_entries */
    NDBRGEN_PAYLOAD *e = enum_db_entries((NDBRGEN_PAYLOAD *)0);
    while (e)
    {
      unsigned int i;
      if (strcasecmp(name, e->namebuf)==0)
        return &(e->pe);
      i = 0;
      while (e->aliaslist[i])
      {
        if (strcasecmp(name, e->aliaslist[i])==0)
          return &(e->pe);
        i++;
      }
      e = enum_db_entries(e); /* get next */
    }
    pe = scan_builtins(name, 0);
    if (pe)
      return pe;
  }
  h_errno = NO_DATA;
  return (NDBRGEN_RETTYPE *)0;
}

NDBRGEN_RETTYPE * getprotobynumber (int scan_proto)
{
  if (scan_proto >= 0 && scan_proto <= 255)
  {
    struct protoent *pe;
    /* begin enumeration by passing NULL to enum_db_entries */
    NDBRGEN_PAYLOAD *e = enum_db_entries((NDBRGEN_PAYLOAD *)0);
    while (e)
    {
      if (scan_proto == e->pe.p_proto)
        return &(e->pe);
      e = enum_db_entries(e); /* get next */
    }
    pe = scan_builtins(((const char *)0), scan_proto);
    if (pe)
      return pe;
  }
  h_errno = NO_DATA;
  return (NDBRGEN_RETTYPE *)0;
}

/* ======================================================================= */

/* This is our entry parser called by the auto-generated code to 
** parse a line in "/etc/protocols". 
*/
static NDBRGEN_PAYLOAD *parser_cb(const char *tokens[], int numtoks, 
                                  NDBRGEN_PAYLOAD *resP)
{
  NDBRGEN_PAYLOAD *e = (NDBRGEN_PAYLOAD *)0;
  if (numtoks == 2)
  { 
    /*  "ip	0	IP"
    **  "icmp	1	ICMP"
    **  "igmp   2	IGMP"
    */
    if (strlen(tokens[0]) < (sizeof(e->namebuf)-1))
    {
      int proto = atoi(tokens[1]);
      if (proto >= 0 && proto <= 255)
      {
        e = resP;
        e->pe.p_name     = strcpy(e->namebuf, tokens[0]);
        e->pe.p_proto    = proto;
        e->pe.p_aliases  = &(e->aliaslist[0]);

        db_add_aliases( &tokens[2], numtoks-2, e->aliasbuf, sizeof(e->aliasbuf),
             &(e->aliaslist[0]), sizeof(e->aliaslist)/sizeof(e->aliaslist[0]));
      }
    }
  }
  return e;
}


/* ======================================================================= */

#if defined(TEST_PROG)
static void dump_ent(const char *label, const struct protoent *e)
{
  if (!e)
    printf("%s: no data!\n", label);
  else
  {
    int i;
    printf ("%s: p_name='%s' p_proto='%d'\n",
           label, e->p_name, e->p_proto );
    for (i = 0; e->p_aliases[i]; i++)
       printf ("        p_aliases[%d]='%s'\n", i, e->p_aliases[i]);
  }
  return;
}

int main (int argc, char *argv[])
{
  if (argc > 1)
  {
    if (strcmp(argv[1],"all")==0)
      argc = 1; /* fall through */
    else 
    {
      if (*argv[1] >= '0' && *argv[1] <= '9')
        dump_ent("bynumb", getprotobynumber( atoi(argv[1]) ));
      else
        dump_ent("byname", getprotobyname( argv[1] ));
    }
  }
  if (argc == 1)
  {
    struct protoent *e;
    while ((e = getprotoent()) != NULL)
      dump_ent("getent", e);
  }
  return (0);
}
#endif /* TEST_PROG */
